from __future__ import annotations

import datetime
import logging
import time
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence, Tuple

from dbt_semantic_interfaces.call_parameter_sets import ParseWhereFilterException
from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilter
from dbt_semantic_interfaces.pretty_print import pformat_big_objects
from dbt_semantic_interfaces.protocols.dimension import DimensionType
from dbt_semantic_interfaces.protocols.metric import MetricType
from dbt_semantic_interfaces.protocols.where_filter import WhereFilter
from dbt_semantic_interfaces.references import (
    DimensionReference,
    EntityReference,
    MetricReference,
    TimeDimensionReference,
)
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.assert_one_arg import assert_exactly_one_arg_set
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.dataflow_plan import ReadSqlSourceNode
from metricflow.dataset.dataset import DataSet
from metricflow.errors.errors import UnableToSatisfyQueryError
from metricflow.filters.time_constraint import TimeRangeConstraint
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow.protocols.query_parameter import (
    GroupByParameter,
    MetricQueryParameter,
    OrderByQueryParameter,
    TimeDimensionQueryParameter,
)
from metricflow.query.query_exceptions import InvalidQueryException
from metricflow.specs.column_assoc import ColumnAssociationResolver
from metricflow.specs.specs import (
    DimensionSpec,
    EntitySpec,
    LinkableInstanceSpec,
    LinkableSpecSet,
    MetricFlowQuerySpec,
    MetricSpec,
    OrderBySpec,
    TimeDimensionSpec,
    WhereFilterSpec,
)
from metricflow.specs.where_filter_transform import WhereSpecFactory
from metricflow.time.date_part import DatePart
from metricflow.time.time_granularity_solver import (
    PartialTimeDimensionSpec,
    RequestTimeGranularityException,
    TimeGranularitySolver,
)

logging.captureWarnings(True)
import rapidfuzz.fuzz  # noqa: E402
import rapidfuzz.process  # noqa: E402

logging.captureWarnings(False)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class QueryTimeLinkableSpecSet:
    """Linkable specs that are specified at query time.

    This is different from LinkableSpecSet in that it allows for partially specified time dimensions. e.g. in a query,
    metric_time might be specified without a granularity suffix, but means the lowest possible time granularity.
    """

    dimension_specs: Tuple[DimensionSpec, ...]
    time_dimension_specs: Tuple[TimeDimensionSpec, ...]
    partial_time_dimension_specs: Tuple[PartialTimeDimensionSpec, ...]
    entity_specs: Tuple[EntitySpec, ...]

    @staticmethod
    def create_from_linkable_spec_set(linkable_spec_set: LinkableSpecSet) -> QueryTimeLinkableSpecSet:  # noqa: D
        return QueryTimeLinkableSpecSet(
            dimension_specs=linkable_spec_set.dimension_specs,
            time_dimension_specs=linkable_spec_set.time_dimension_specs,
            partial_time_dimension_specs=(),
            entity_specs=linkable_spec_set.entity_specs,
        )

    @staticmethod
    def combine(spec_sets: Sequence[QueryTimeLinkableSpecSet]) -> QueryTimeLinkableSpecSet:  # noqa: D
        return QueryTimeLinkableSpecSet(
            dimension_specs=tuple(
                dimension_spec for spec_set in spec_sets for dimension_spec in spec_set.dimension_specs
            ),
            time_dimension_specs=tuple(
                time_dimension_spec for spec_set in spec_sets for time_dimension_spec in spec_set.time_dimension_specs
            ),
            partial_time_dimension_specs=tuple(
                partial_time_dimension_spec
                for spec_set in spec_sets
                for partial_time_dimension_spec in spec_set.partial_time_dimension_specs
            ),
            entity_specs=tuple(entity_spec for spec_set in spec_sets for entity_spec in spec_set.entity_specs),
        )


class MetricFlowQueryParser:
    """Parse input strings from the user into a metric query specification.

    Definitions:
    element name - the name of an element (measure, dimension, entity) in a semantic model, or a metric name.
    qualified name - an element name with prefixes and suffixes added to it that further describe transformations or
    conditions for the element to retrieve. e.g. "ds__month" is the "ds" time dimension at the "month" granularity. Or
    "user_id__country" is the "country" dimension that is retrieved by joining "user_id" to the measure semantic model.
    """

    def __init__(  # noqa: D
        self,
        column_association_resolver: ColumnAssociationResolver,
        model: SemanticManifestLookup,
        read_nodes: Sequence[ReadSqlSourceNode],
        node_output_resolver: DataflowPlanNodeOutputDataSetResolver,
    ) -> None:
        self._column_association_resolver = column_association_resolver
        self._model = model
        self._metric_lookup = model.metric_lookup
        self._semantic_model_lookup = model.semantic_model_lookup
        self._node_output_resolver = node_output_resolver
        self._read_nodes = read_nodes

        # Set up containers for known element names
        self._known_entity_element_references = self._semantic_model_lookup.get_entity_references()

        self._known_time_dimension_element_references = [DataSet.metric_time_dimension_reference()]
        self._known_dimension_element_references = []
        for dimension_reference in self._semantic_model_lookup.get_dimension_references():
            dimension = self._semantic_model_lookup.get_dimension(dimension_reference)
            if dimension.type == DimensionType.CATEGORICAL:
                self._known_dimension_element_references.append(dimension_reference)
            elif dimension.type == DimensionType.TIME:
                self._known_time_dimension_element_references.append(dimension_reference.time_dimension_reference)
            else:
                raise RuntimeError(f"Unhandled linkable type: {dimension.type}")

        self._known_metric_names = set(self._metric_lookup.metric_references)
        self._metric_time_dimension_reference = DataSet.metric_time_dimension_reference()
        self._time_granularity_solver = TimeGranularitySolver(
            semantic_manifest_lookup=self._model,
            read_nodes=self._read_nodes,
            node_output_resolver=self._node_output_resolver,
        )

    @staticmethod
    def _top_fuzzy_matches(
        item: str,
        candidate_items: Sequence[str],
        max_suggestions: int = 6,
        min_score: int = 50,
    ) -> Sequence[str]:
        """Return the top items (by edit distance) in candidate_items that fuzzy matches the given item."""
        # Rank choices by edit distance score.
        # extract() return a tuple like (name, score)
        top_ranked_suggestions = sorted(
            rapidfuzz.process.extract(
                # This scorer seems to return the best results.
                item,
                list(candidate_items),
                limit=max_suggestions,
                scorer=rapidfuzz.fuzz.token_set_ratio,
            ),
            # Put the highest scoring item at the top of the list.
            key=lambda x: x[1],
            reverse=True,
        )
        # If the scores are too low, then it looks nonsensical, so remove those.
        top_ranked_suggestions = [x for x in top_ranked_suggestions if x[1] > min_score]
        return [x[0] for x in top_ranked_suggestions]

    def parse_and_validate_query(
        self,
        metric_names: Optional[Sequence[str]] = None,
        metrics: Optional[Sequence[MetricQueryParameter]] = None,
        group_by_names: Optional[Sequence[str]] = None,
        group_by: Optional[Tuple[GroupByParameter, ...]] = None,
        limit: Optional[int] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
        where_constraint: Optional[WhereFilter] = None,
        where_constraint_str: Optional[str] = None,
        order_by_names: Optional[Sequence[str]] = None,
        order_by: Optional[Sequence[OrderByQueryParameter]] = None,
    ) -> MetricFlowQuerySpec:
        """Parse the query into spec objects, validating them in the process.

        e.g. make sure that the given metric is a valid metric name.
        """
        start_time = time.time()
        try:
            return self._parse_and_validate_query(
                metric_names=metric_names,
                metrics=metrics,
                group_by_names=group_by_names,
                group_by=group_by,
                limit=limit,
                time_constraint_start=time_constraint_start,
                time_constraint_end=time_constraint_end,
                where_constraint=where_constraint,
                where_constraint_str=where_constraint_str,
                order_by_names=order_by_names,
                order_by=order_by,
            )
        finally:
            logger.info(f"Parsing the query took: {time.time() - start_time:.2f}s")

    def _validate_no_time_dimension_query(self, metric_references: Sequence[MetricReference]) -> None:
        """Validate if all requested metrics are queryable without a time dimension."""
        for metric_reference in metric_references:
            metric = self._metric_lookup.get_metric(metric_reference)
            if metric.type == MetricType.CUMULATIVE:
                # Cumulative metrics configured with a window/grain_to_date cannot be queried without a dimension.
                if metric.type_params.window or metric.type_params.grain_to_date:
                    raise UnableToSatisfyQueryError(
                        f"Metric {metric.name} is a cumulative metric specified with a window/grain_to_date "
                        f"which must be queried with the dimension 'metric_time'.",
                    )
            elif metric.type == MetricType.DERIVED:
                for input_metric in metric.type_params.metrics or []:
                    if input_metric.offset_window or input_metric.offset_to_grain:
                        raise UnableToSatisfyQueryError(
                            f"Metric '{metric.name}' is a derived metric that contains input metrics with "
                            "an `offset_window` or `offset_to_grain` which must be queried with the "
                            "dimension 'metric_time'."
                        )

    def _validate_linkable_specs_for_metrics(
        self,
        metric_references: Tuple[MetricReference, ...],
        all_linkable_specs: QueryTimeLinkableSpecSet,
        time_dimension_specs: Tuple[TimeDimensionSpec, ...],
    ) -> None:
        if not metric_references:
            return None

        invalid_group_bys = self._get_invalid_linkable_specs(
            metric_references=metric_references,
            dimension_specs=all_linkable_specs.dimension_specs,
            time_dimension_specs=time_dimension_specs,
            entity_specs=all_linkable_specs.entity_specs,
        )
        if len(invalid_group_bys) > 0:
            valid_group_by_names_for_metrics = sorted(
                x.qualified_name
                for x in self._metric_lookup.element_specs_for_metrics(metric_references=list(metric_references))
            )
            # Create suggestions for invalid dimensions in case the user made a typo.
            suggestion_sections = {}
            for invalid_group_by in invalid_group_bys:
                suggestions = MetricFlowQueryParser._top_fuzzy_matches(
                    item=invalid_group_by.qualified_name,
                    candidate_items=valid_group_by_names_for_metrics,
                )
                section_key = f"Suggestions for invalid dimension '{invalid_group_by.qualified_name}'"
                section_value = pformat_big_objects(suggestions)
                suggestion_sections[section_key] = section_value
            raise UnableToSatisfyQueryError(
                f"Dimensions {[x.qualified_name for x in invalid_group_bys]} cannot be "
                f"resolved for metrics {[x.element_name for x in metric_references]}. The invalid dimension may not "
                f"exist, require an ambiguous join (e.g. a join path that can be satisfied in multiple ways), "
                f"or require a fanout join.",
                context=suggestion_sections,
            )

    def _construct_metric_specs_for_query(
        self, metric_references: Tuple[MetricReference, ...]
    ) -> Tuple[MetricSpec, ...]:
        """Populate MetricSpecs.

        Construct MetricSpecs in the preaggregated state to pass into the DataflowPlanBuilder.

        NOTE: Currently, we populate only the metrics provided in the query, but with derived metrics,
        there are nested metrics that do not get populated here, but rather in the MetricSemantic during
        the builder process. This is a process that should be refined altogther.
        """
        metric_specs = []
        for metric_reference in metric_references:
            metric = self._metric_lookup.get_metric(metric_reference)
            metric_where_constraint: Optional[WhereFilterSpec] = None
            if metric.filter:
                # add constraint to MetricSpec
                metric_where_constraint = WhereSpecFactory(
                    column_association_resolver=self._column_association_resolver,
                ).create_from_where_filter_intersection(metric.filter)
            # TODO: Directly initializing Spec object instead of using a factory method since
            #       importing WhereConstraintConverter is a problem in specs.py
            metric_specs.append(
                MetricSpec(
                    element_name=metric_reference.element_name,
                    constraint=metric_where_constraint,
                )
            )
        return tuple(metric_specs)

    def _get_metric_names(
        self, metric_names: Optional[Sequence[str]], metrics: Optional[Sequence[MetricQueryParameter]]
    ) -> Sequence[str]:
        if not (metric_names or metrics):
            return []

        assert_exactly_one_arg_set(metric_names=metric_names, metrics=metrics)
        return metric_names if metric_names else [m.name for m in metrics] if metrics else []

    def _get_where_filter(
        self,
        where_constraint: Optional[WhereFilter] = None,
        where_constraint_str: Optional[str] = None,
    ) -> Optional[WhereFilter]:
        assert not (
            where_constraint and where_constraint_str
        ), "Both where_constraint and where_constraint_str were set, but if a where is specified you should only use one of these!"
        return (
            PydanticWhereFilter(where_sql_template=where_constraint_str) if where_constraint_str else where_constraint
        )

    def _parse_and_validate_query(
        self,
        metric_names: Optional[Sequence[str]] = None,
        metrics: Optional[Sequence[MetricQueryParameter]] = None,
        group_by_names: Optional[Sequence[str]] = None,
        group_by: Optional[Tuple[GroupByParameter, ...]] = None,
        limit: Optional[int] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
        where_constraint: Optional[WhereFilter] = None,
        where_constraint_str: Optional[str] = None,
        order_by_names: Optional[Sequence[str]] = None,
        order_by: Optional[Sequence[OrderByQueryParameter]] = None,
    ) -> MetricFlowQuerySpec:
        metric_names = self._get_metric_names(metric_names, metrics)
        where_filter = self._get_where_filter(where_constraint, where_constraint_str)

        # Get metric references used for validations
        # In a case of derived metric, all the input metrics would be here.
        metric_references = self._parse_metric_names(metric_names)
        if time_constraint_start is None:
            time_constraint_start = TimeRangeConstraint.ALL_TIME_BEGIN()
        elif time_constraint_start < TimeRangeConstraint.ALL_TIME_BEGIN():
            logger.warning(
                f"Start time for the supplied time constraint {time_constraint_start.isoformat()} is <less> than the "
                f"minimum allowed ('{TimeRangeConstraint.ALL_TIME_BEGIN().isoformat()}'). Changing to the minimum."
            )
            time_constraint_start = TimeRangeConstraint.ALL_TIME_BEGIN()
        elif time_constraint_start > TimeRangeConstraint.ALL_TIME_END():
            logger.warning(
                f"Start time for the supplied time constraint {time_constraint_start.isoformat()} is > than the "
                f"maximum allowed ('{TimeRangeConstraint.ALL_TIME_END().isoformat()}'). Changing to the maximum."
            )
            time_constraint_start = TimeRangeConstraint.ALL_TIME_END()

        if time_constraint_end is None:
            time_constraint_end = TimeRangeConstraint.ALL_TIME_END()
        elif time_constraint_end > TimeRangeConstraint.ALL_TIME_END():
            logger.warning(
                f"End time for the supplied time constraint {time_constraint_end.isoformat()} is > than the "
                f"maximum allowed ('{TimeRangeConstraint.ALL_TIME_END().isoformat()}'). Changing to the maximum."
            )
            time_constraint_end = TimeRangeConstraint.ALL_TIME_END()
        elif time_constraint_end < TimeRangeConstraint.ALL_TIME_BEGIN():
            logger.warning(
                f"End time for the supplied time constraint {time_constraint_end.isoformat()} is less than the "
                f"minimum allowed ('{TimeRangeConstraint.ALL_TIME_BEGIN().isoformat()}'). Changing to the minimum."
            )
            time_constraint_end = TimeRangeConstraint.ALL_TIME_BEGIN()

        if time_constraint_end < time_constraint_start:
            raise ValueError(
                f"End of time constraint {time_constraint_end.isoformat()} is before start of time constraint "
                f"{time_constraint_start.isoformat()}"
            )

        time_constraint: Optional[TimeRangeConstraint] = TimeRangeConstraint(
            start_time=time_constraint_start, end_time=time_constraint_end
        )
        if time_constraint == TimeRangeConstraint.all_time():
            # If the time constraint is all time, just ignore and not render
            time_constraint = None

        requested_linkable_specs = self._parse_group_by(
            group_by_names=group_by_names, group_by=group_by, metric_references=metric_references
        )
        where_filter_spec: Optional[WhereFilterSpec] = None
        if where_filter is not None:
            try:
                where_filter_spec = WhereSpecFactory(
                    column_association_resolver=self._column_association_resolver,
                ).create_from_where_filter(where_filter)
            except ParseWhereFilterException as e:
                raise InvalidQueryException(
                    f"Error parsing the where filter: {where_filter.where_sql_template}. {e}"
                ) from e

            where_spec_set = QueryTimeLinkableSpecSet.create_from_linkable_spec_set(where_filter_spec.linkable_spec_set)
            requested_linkable_specs_with_requested_filter_specs = QueryTimeLinkableSpecSet.combine(
                (
                    requested_linkable_specs,
                    where_spec_set,
                )
            )
        else:
            requested_linkable_specs_with_requested_filter_specs = requested_linkable_specs

        partial_time_dimension_spec_replacements = (
            self._time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
                metric_references=metric_references,
                partial_time_dimension_specs=requested_linkable_specs.partial_time_dimension_specs,
            )
        )

        time_dimension_specs = requested_linkable_specs.time_dimension_specs + tuple(
            time_dimension_spec for _, time_dimension_spec in partial_time_dimension_spec_replacements.items()
        )
        if len(time_dimension_specs) == 0:
            self._validate_no_time_dimension_query(metric_references=metric_references)

        self._time_granularity_solver.validate_time_granularity(metric_references, time_dimension_specs)
        self._validate_date_part(metric_references, time_dimension_specs)

        order_by_specs = self._parse_order_by(
            order_by_names=order_by_names,
            order_by=order_by,
            time_dimension_spec_replacements=partial_time_dimension_spec_replacements,
        )

        # For each metric, verify that it's possible to retrieve all group by elements, including the ones as required
        # by the filters.
        # TODO: Consider moving this logic into _validate_linkable_specs_for_metrics().
        for metric_reference in metric_references:
            metric = self._metric_lookup.get_metric(metric_reference)
            if metric.filter is not None:
                group_by_specs_for_one_metric = self._parse_group_by(
                    group_by_names=group_by_names,
                    group_by=group_by,
                    metric_references=(metric_reference,),
                )

                # Combine the group by elements from the query with the group by elements that are required by the
                # metric filter to see if that's a valid set that could be queried.
                where_spec = WhereSpecFactory(
                    column_association_resolver=self._column_association_resolver
                ).create_from_where_filter_intersection(metric.filter)
                self._validate_linkable_specs_for_metrics(
                    metric_references=(metric_reference,),
                    all_linkable_specs=(
                        QueryTimeLinkableSpecSet.combine(
                            (
                                group_by_specs_for_one_metric,
                                QueryTimeLinkableSpecSet.create_from_linkable_spec_set(where_spec.linkable_spec_set),
                            )
                        )
                        if where_spec is not None
                        else group_by_specs_for_one_metric
                    ),
                    time_dimension_specs=time_dimension_specs,
                )

        # Validate all of them together.
        self._validate_linkable_specs_for_metrics(
            metric_references=metric_references,
            all_linkable_specs=requested_linkable_specs_with_requested_filter_specs,
            time_dimension_specs=time_dimension_specs,
        )

        self._validate_order_by_specs(
            order_by_specs=order_by_specs,
            metric_references=metric_references,
            linkable_specs=LinkableSpecSet(
                dimension_specs=requested_linkable_specs.dimension_specs,
                time_dimension_specs=time_dimension_specs,
                entity_specs=requested_linkable_specs.entity_specs,
            ),
        )

        # Update constraints to be appropriate for the time granularity of the query.
        if time_constraint:
            logger.info(f"Time constraint before adjustment is {time_constraint}")
            time_constraint = self._adjust_time_range_constraint(
                metric_references=metric_references,
                time_dimension_specs=time_dimension_specs,
                time_range_constraint=time_constraint,
            )
            logger.info(f"Time constraint after adjustment is {time_constraint}")

        if limit is not None and limit < 0:
            raise InvalidQueryException(f"Limit was specified as {limit}, which is < 0.")

        if where_filter_spec:
            self._time_granularity_solver.validate_time_granularity(
                metric_references=metric_references,
                time_dimension_specs=where_filter_spec.linkable_spec_set.time_dimension_specs,
            )

        base_metric_references = self._parse_metric_names(metric_names, traverse_metric_inputs=False)
        metric_specs = self._construct_metric_specs_for_query(base_metric_references)

        return MetricFlowQuerySpec(
            metric_specs=metric_specs,
            dimension_specs=requested_linkable_specs.dimension_specs,
            entity_specs=requested_linkable_specs.entity_specs,
            time_dimension_specs=time_dimension_specs,
            order_by_specs=order_by_specs,
            time_range_constraint=time_constraint,
            where_constraint=where_filter_spec,
            limit=limit,
        )

    def _validate_order_by_specs(
        self,
        order_by_specs: Sequence[OrderBySpec],
        metric_references: Sequence[MetricReference],
        linkable_specs: LinkableSpecSet,
    ) -> None:
        """Checks that the order by specs references an item in the query."""
        # TODO: this is a workaround
        # Need to figure out whether we should clean up OrderBySpec or if we have to actually pass a fully resolved MetricSpec here
        metric_specs = [MetricSpec.from_reference(metric_reference) for metric_reference in metric_references]
        for order_by_spec in order_by_specs:
            if not (
                order_by_spec.instance_spec in metric_specs
                or order_by_spec.instance_spec in linkable_specs.dimension_specs
                or order_by_spec.instance_spec in linkable_specs.time_dimension_specs
                or order_by_spec.instance_spec in linkable_specs.entity_specs
            ):
                raise InvalidQueryException(f"Order by item {order_by_spec} not in the query")

    def _validate_date_part(
        self, metric_references: Sequence[MetricReference], time_dimension_specs: Sequence[TimeDimensionSpec]
    ) -> None:
        """Validate that date parts can be used for metrics.

        TODO: figure out expected behavior for date part with these types of metrics.
        """
        date_part_requested = False
        for time_dimension_spec in time_dimension_specs:
            if time_dimension_spec.date_part:
                date_part_requested = True
                if time_dimension_spec.date_part.to_int() < time_dimension_spec.time_granularity.to_int():
                    raise RequestTimeGranularityException(
                        f"Date part {time_dimension_spec.date_part.name} is not compatible with time granularity "
                        f"{time_dimension_spec.time_granularity.name}. Compatible granularities include: "
                        f"{[granularity.name for granularity in time_dimension_spec.date_part.compatible_granularities]}"
                    )
        if date_part_requested:
            for metric_reference in metric_references:
                metric = self._metric_lookup.get_metric(metric_reference)
                if metric.type == MetricType.CUMULATIVE:
                    raise UnableToSatisfyQueryError("Cannot extract date part for cumulative metrics.")
                elif metric.type == MetricType.DERIVED:
                    for input_metric in metric.type_params.metrics or []:
                        if input_metric.offset_to_grain:
                            raise UnableToSatisfyQueryError(
                                "Cannot extract date part for metrics with offset_to_grain."
                            )

    def _adjust_time_range_constraint(
        self,
        metric_references: Sequence[MetricReference],
        time_dimension_specs: Sequence[TimeDimensionSpec],
        time_range_constraint: TimeRangeConstraint,
    ) -> TimeRangeConstraint:
        """Adjust the time range constraint so that it matches the boundaries of the granularity of the result."""
        self._time_granularity_solver.validate_time_granularity(metric_references, time_dimension_specs)

        smallest_primary_time_granularity_in_query = self._find_smallest_metric_time_dimension_spec_granularity(
            time_dimension_specs
        )
        if smallest_primary_time_granularity_in_query is not None:
            adjust_to_granularity = smallest_primary_time_granularity_in_query

        else:
            partial_metric_time_spec = PartialTimeDimensionSpec(
                entity_links=(),
                element_name=DataSet.metric_time_dimension_name(),
            )
            partial_time_dimension_spec_to_time_dimension_spec = (
                self._time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
                    metric_references=metric_references, partial_time_dimension_specs=(partial_metric_time_spec,)
                )
            )
            adjust_to_granularity = partial_time_dimension_spec_to_time_dimension_spec[
                partial_metric_time_spec
            ].time_granularity
        logger.info(f"Adjusted metric time granularity is {adjust_to_granularity}")
        return self._time_granularity_solver.adjust_time_range_to_granularity(
            time_range_constraint, adjust_to_granularity
        )

    def _find_replacement_for_metric_time_dimension(
        self, replacements: Dict[PartialTimeDimensionSpec, TimeDimensionSpec]
    ) -> Tuple[PartialTimeDimensionSpec, TimeDimensionSpec]:
        for partial_time_dimension_spec_to_replace, replace_with_time_dimension_spec in replacements.items():
            if (
                partial_time_dimension_spec_to_replace.element_name
                == self._metric_time_dimension_reference.element_name
                and partial_time_dimension_spec_to_replace.entity_links == ()
            ):
                return (
                    partial_time_dimension_spec_to_replace,
                    replace_with_time_dimension_spec,
                )

        raise RuntimeError(f"Replacement for metric time dimension '{self._metric_time_dimension_reference}' not found")

    def _metric_time_dimension_specified_without_granularity(
        self, partial_time_dimension_specs: Sequence[PartialTimeDimensionSpec]
    ) -> bool:
        # This detects a user query like: "query --metrics monthly_bookings --dimensions ds"
        # The granularity for "ds" is not specified by the user.
        for time_dimension_spec in partial_time_dimension_specs:
            if (
                time_dimension_spec.element_name == self._metric_time_dimension_reference.element_name
                and time_dimension_spec.entity_links == ()
            ):
                return True
        return False

    def _find_smallest_metric_time_dimension_spec_granularity(
        self, time_dimension_specs: Sequence[TimeDimensionSpec]
    ) -> Optional[TimeGranularity]:
        metric_time_dimension_specs: List[TimeDimensionSpec] = [
            x
            for x in time_dimension_specs
            if (
                x.element_name == self._metric_time_dimension_reference.element_name
                and x.entity_links == ()
                and x.time_granularity
            )
        ]

        metric_time_dimension_specs.sort(key=lambda x: x.time_granularity.to_int())
        if len(metric_time_dimension_specs) > 0:
            return metric_time_dimension_specs[0].time_granularity
        else:
            return None

    def _parse_metric_names(
        self, metric_names: Sequence[str], traverse_metric_inputs: bool = True
    ) -> Tuple[MetricReference, ...]:
        """Converts metric names into metric names. An exception is thrown if the name is invalid."""
        # The config must be lower-case, so we lower case for case-insensitivity against query inputs from the user.
        metric_names = [x.lower() for x in metric_names]

        known_metric_names = set(self._metric_lookup.metric_references)
        metric_references: List[MetricReference] = []
        for metric_name in metric_names:
            metric_reference = MetricReference(element_name=metric_name)
            if metric_reference not in known_metric_names:
                suggestions = {
                    f"Suggestions for '{metric_name}'": pformat_big_objects(
                        MetricFlowQueryParser._top_fuzzy_matches(
                            item=metric_name,
                            candidate_items=[x.element_name for x in self._metric_lookup.metric_references],
                        )
                    )
                }
                raise UnableToSatisfyQueryError(
                    f"Unknown metric: '{metric_name}'",
                    context=suggestions,
                )
            metric_references.append(metric_reference)
            if traverse_metric_inputs:
                metric = self._metric_lookup.get_metric(metric_reference)
                if metric.type == MetricType.DERIVED:
                    input_metrics = self._parse_metric_names([metric.name for metric in metric.input_metrics])
                    metric_references.extend(list(input_metrics))
        return tuple(metric_references)

    def _parse_group_by(
        self,
        metric_references: Sequence[MetricReference],
        group_by_names: Optional[Sequence[str]] = None,
        group_by: Optional[Tuple[GroupByParameter, ...]] = None,
    ) -> QueryTimeLinkableSpecSet:
        """Convert the linkable spec names into the respective specification objects."""
        assert not (
            group_by_names and group_by
        ), "Both group_by_names and group_by were set, but if a group by is specified you should only use one of these!"

        structured_names: List[StructuredLinkableSpecName] = []
        if group_by_names:
            group_by_names = [x.lower() for x in group_by_names]
            structured_names = [StructuredLinkableSpecName.from_name(name) for name in group_by_names]
        elif group_by:
            for group_by_obj in group_by:
                parsed_name = StructuredLinkableSpecName.from_name(group_by_obj.name.lower())
                structured_name = StructuredLinkableSpecName(
                    entity_link_names=parsed_name.entity_link_names,
                    element_name=parsed_name.element_name,
                    time_granularity=group_by_obj.grain
                    if isinstance(group_by_obj, TimeDimensionQueryParameter)
                    else None,
                    date_part=group_by_obj.date_part if isinstance(group_by_obj, TimeDimensionQueryParameter) else None,
                )
                structured_names.append(structured_name)

        dimension_specs = []
        time_dimension_specs = []
        partial_time_dimension_specs = []
        entity_specs = []

        for structured_name in structured_names:
            element_name = structured_name.element_name
            entity_links = tuple(EntityReference(element_name=x) for x in structured_name.entity_link_names)
            # Create the spec based on the type of element referenced.
            if TimeDimensionReference(element_name=element_name) in self._known_time_dimension_element_references:
                if structured_name.time_granularity and not structured_name.date_part:
                    time_dimension_specs.append(
                        TimeDimensionSpec(
                            element_name=element_name,
                            entity_links=entity_links,
                            time_granularity=structured_name.time_granularity,
                        )
                    )
                else:
                    partial_time_dimension_spec = PartialTimeDimensionSpec(
                        element_name=element_name, entity_links=entity_links, date_part=structured_name.date_part
                    )
                    # If both granularity & date_part are requested, verify requested & resolved granularities match.
                    if structured_name.time_granularity and structured_name.date_part:
                        self._verify_resolved_granularity_for_date_part(
                            requested_dimension_structured_name=structured_name,
                            partial_time_dimension_spec=partial_time_dimension_spec,
                            metric_references=metric_references,
                        )
                    partial_time_dimension_specs.append(partial_time_dimension_spec)

            elif DimensionReference(element_name=element_name) in self._known_dimension_element_references:
                dimension_specs.append(DimensionSpec(element_name=element_name, entity_links=entity_links))
            elif EntityReference(element_name=element_name) in self._known_entity_element_references:
                entity_specs.append(EntitySpec(element_name=element_name, entity_links=entity_links))
            else:
                valid_group_bys_for_metrics = self._metric_lookup.element_specs_for_metrics(list(metric_references))
                valid_group_by_names_for_metrics = sorted(
                    list(
                        set(x.qualified_name if group_by_names else x.element_name for x in valid_group_bys_for_metrics)
                    )
                )

                # If requested by name, show qualified name. If requested as object, show element name.
                display_name = structured_name.qualified_name if group_by_names else element_name
                suggestions = {
                    f"Suggestions for '{display_name}'": pformat_big_objects(
                        MetricFlowQueryParser._top_fuzzy_matches(
                            item=display_name,
                            candidate_items=valid_group_by_names_for_metrics,
                        )
                    )
                }
                raise UnableToSatisfyQueryError(
                    f"Unknown element name '{element_name}' in dimension name '{display_name}'"
                    if group_by_names
                    else f"Unknown dimension {element_name}",
                    context=suggestions,
                )

        return QueryTimeLinkableSpecSet(
            dimension_specs=tuple(dimension_specs),
            time_dimension_specs=tuple(time_dimension_specs),
            partial_time_dimension_specs=tuple(partial_time_dimension_specs),
            entity_specs=tuple(entity_specs),
        )

    def _verify_resolved_granularity_for_date_part(
        self,
        requested_dimension_structured_name: StructuredLinkableSpecName,
        partial_time_dimension_spec: PartialTimeDimensionSpec,
        metric_references: Sequence[MetricReference],
    ) -> None:
        """Enforce that any granularity value associated with a date part query is the minimum.

        By default, we will always ensure that a date_part query request uses the minimum granularity.
        However, there are some interfaces where the user must pass in a granularity, so we need a check to
        ensure that the correct value was passed in.
        """
        resolved_granularity = self._time_granularity_solver.find_minimum_granularity_for_partial_time_dimension_spec(
            partial_time_dimension_spec=partial_time_dimension_spec, metric_references=metric_references
        )
        if resolved_granularity != requested_dimension_structured_name.time_granularity:
            raise RequestTimeGranularityException(
                f"When applying a date part to dimension '{requested_dimension_structured_name.qualified_name}' with "
                f"metrics {[metric.element_name for metric in metric_references]}, only {resolved_granularity.name} "
                "granularity can be used."
            )

    def _get_invalid_linkable_specs(
        self,
        metric_references: Tuple[MetricReference, ...],
        dimension_specs: Tuple[DimensionSpec, ...],
        time_dimension_specs: Tuple[TimeDimensionSpec, ...],
        entity_specs: Tuple[EntitySpec, ...],
    ) -> List[LinkableInstanceSpec]:
        """Checks that each requested linkable instance can be retrieved for the given metric."""
        invalid_linkable_specs: List[LinkableInstanceSpec] = []
        # TODO: distinguish between dimensions that invalid via typo vs ambiguous join path
        valid_linkable_specs = self._metric_lookup.element_specs_for_metrics(metric_references=list(metric_references))

        for dimension_spec in dimension_specs:
            if dimension_spec not in valid_linkable_specs:
                invalid_linkable_specs.append(dimension_spec)

        for entity_spec in entity_specs:
            if entity_spec not in valid_linkable_specs:
                invalid_linkable_specs.append(entity_spec)

        for time_dimension_spec in time_dimension_specs:
            time_dimension_spec_without_date_part = time_dimension_spec
            if time_dimension_spec.date_part:
                # TODO: remove this workaround & add date_part specs to validation paths.
                time_dimension_spec_without_date_part = TimeDimensionSpec(
                    element_name=time_dimension_spec.element_name,
                    entity_links=time_dimension_spec.entity_links,
                    time_granularity=time_dimension_spec.time_granularity,
                    aggregation_state=time_dimension_spec.aggregation_state,
                )
            if (
                time_dimension_spec_without_date_part not in valid_linkable_specs
                # Because the metric time dimension is a virtual dimension that's not in the model, it won't be included
                # in valid_linkable_specs.
                and time_dimension_spec.reference != DataSet.metric_time_dimension_reference()
            ):
                invalid_linkable_specs.append(time_dimension_spec)

        return invalid_linkable_specs

    def _parse_order_by(
        self,
        time_dimension_spec_replacements: Dict[PartialTimeDimensionSpec, TimeDimensionSpec],
        order_by_names: Optional[Sequence[str]] = None,
        order_by: Optional[Sequence[OrderByQueryParameter]] = None,
    ) -> Tuple[OrderBySpec, ...]:
        """time_dimension_spec_replacements is used to replace a partial spec from parsing the names to a full one."""
        # TODO: Validate entity links
        # TODO: Validate order by items are in the query

        assert not (
            order_by_names and order_by
        ), "Both order_by_names and order_by were set, but if an order by is specified you should only use one of these!"

        order_by_specs: List[OrderBySpec] = []
        for order in order_by_names or order_by or []:
            # Order might be string or object - parse into relevant parts.
            time_granularity: Optional[TimeGranularity] = None
            date_part: Optional[DatePart] = None
            if isinstance(order, str):
                order = order.lower()
                # Note: date part cannot be requested via string parameter.
                descending = False
                if order.startswith("-"):
                    order = order[1:]
                    descending = True
                parsed_name = StructuredLinkableSpecName.from_name(order)
                time_granularity = parsed_name.time_granularity
            else:
                descending = order.descending
                parsed_name = StructuredLinkableSpecName.from_name(order.order_by.name.lower())
                if isinstance(order.order_by, TimeDimensionQueryParameter):
                    time_granularity = order.order_by.grain
                    date_part = order.order_by.date_part
                if parsed_name.time_granularity and parsed_name.time_granularity != time_granularity:
                    raise InvalidQueryException("Must use object syntax to request a time granularity.")

            if MetricReference(element_name=parsed_name.element_name) in self._known_metric_names:
                if time_granularity:
                    raise InvalidQueryException(
                        f"Order by item '{order}' references a metric but has a time granularity"
                    )
                if parsed_name.entity_link_names:
                    raise InvalidQueryException(f"Order by item '{order}' references a metric but has entity links")
                order_by_specs.append(
                    OrderBySpec(
                        instance_spec=MetricSpec(element_name=parsed_name.element_name),
                        descending=descending,
                    )
                )
            elif DimensionReference(element_name=parsed_name.element_name) in self._known_dimension_element_references:
                if time_granularity:
                    raise InvalidQueryException(
                        f"Order by item '{order}' references a categorical dimension but has a time " f"granularity"
                    )
                order_by_specs.append(
                    OrderBySpec(
                        instance_spec=DimensionSpec(
                            element_name=parsed_name.element_name,
                            entity_links=tuple(EntityReference(element_name=x) for x in parsed_name.entity_link_names),
                        ),
                        descending=descending,
                    )
                )
            elif (
                TimeDimensionReference(element_name=parsed_name.element_name)
                in self._known_time_dimension_element_references
            ):
                entity_links = tuple(EntityReference(element_name=x) for x in parsed_name.entity_link_names)
                if time_granularity:
                    order_by_specs.append(
                        OrderBySpec(
                            instance_spec=TimeDimensionSpec(
                                element_name=parsed_name.element_name,
                                entity_links=entity_links,
                                time_granularity=time_granularity,
                                date_part=date_part,
                            ),
                            descending=descending,
                        )
                    )
                else:
                    # If the time granularity for an order by wasn't specified, replace it with the same time
                    # granularity as it was done for the requested dimensions.
                    partial_time_dimension_spec = PartialTimeDimensionSpec(
                        element_name=parsed_name.element_name,
                        entity_links=entity_links,
                        date_part=date_part,
                    )

                    if partial_time_dimension_spec in time_dimension_spec_replacements:
                        order_by_specs.append(
                            OrderBySpec(
                                instance_spec=time_dimension_spec_replacements[partial_time_dimension_spec],
                                descending=descending,
                            )
                        )
                    else:
                        raise RequestTimeGranularityException(
                            f"Order by item '{order}' does not specify a time granularity and it does not "
                            f"match a requested time dimension"
                        )

            elif EntityReference(element_name=parsed_name.element_name) in self._known_entity_element_references:
                if time_granularity:
                    raise InvalidQueryException(
                        f"Order by item '{order}' references an entity but has a time granularity"
                    )
                order_by_specs.append(
                    OrderBySpec(
                        instance_spec=EntitySpec(
                            element_name=parsed_name.element_name,
                            entity_links=tuple(EntityReference(element_name=x) for x in parsed_name.entity_link_names),
                        ),
                        descending=descending,
                    )
                )
            else:
                raise InvalidQueryException(f"Order by item '{order}' references an element that is not known")

        return tuple(order_by_specs)



"""Protocols for classes meant to manage semantic information of useful object types.

These are useful as more generic descriptors of interfaces around classes like semantic containers and such, which
might be best used in a read-only mode. These containers could also pull in extra dependencies in their internal
implementations, dependencies which are best isolated inside the packages containing the concrete objects rather
than the interface specifications.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict, FrozenSet, Optional, Sequence, Set

from dbt_semantic_interfaces.protocols.dimension import Dimension
from dbt_semantic_interfaces.protocols.entity import Entity
from dbt_semantic_interfaces.protocols.measure import Measure
from dbt_semantic_interfaces.protocols.metric import Metric
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    DimensionReference,
    EntityReference,
    MeasureReference,
    MetricReference,
    SemanticModelElementReference,
    SemanticModelReference,
    TimeDimensionReference,
)

from metricflow.model.semantics.element_group import ElementGrouper
from metricflow.model.semantics.linkable_element_properties import LinkableElementProperties
from metricflow.specs.column_assoc import ColumnAssociationResolver
from metricflow.specs.specs import (
    LinkableInstanceSpec,
    MeasureSpec,
    MetricInputMeasureSpec,
    MetricSpec,
    NonAdditiveDimensionSpec,
)


class SemanticModelAccessor(ABC):
    """Interface for accessing semantic information about a set of semantic model objects.

    This is primarily useful for restricting caller access to the subset of container methods and imports we want
    them to use. For example, the SemanticModelLookup class might implement this protocol but also include some
    public methods for adding or removing semantic models from the container, while this protocol only allows the
    caller to invoke the accessor methods which retrieve semantic information about the collected semantic models.
    """

    @abstractmethod
    def get_dimension_references(self) -> Sequence[DimensionReference]:
        """Retrieve all dimension references from the collection of semantic models."""
        raise NotImplementedError

    @abstractmethod
    def get_dimension(self, dimension_reference: DimensionReference) -> Dimension:
        """Retrieve the dimension model object associated with the time dimension reference."""
        raise NotImplementedError

    @abstractmethod
    def get_time_dimension(self, time_dimension_reference: TimeDimensionReference) -> Dimension:
        """Retrieve the dimension model object associated with the time dimension reference."""
        raise NotImplementedError

    @property
    @abstractmethod
    def measure_references(self) -> Sequence[MeasureReference]:
        """Return all measure references from the collection of semantic models."""
        raise NotImplementedError

    @property
    @abstractmethod
    def non_additive_dimension_specs_by_measure(self) -> Dict[MeasureReference, NonAdditiveDimensionSpec]:
        """Return a mapping from all semi-additive measures to their corresponding non additive dimension parameters.

        This includes all measures with non-additive dimension parameters, if any, from the collection of semantic models.
        """
        raise NotImplementedError

    @abstractmethod
    def get_measure(self, measure_reference: MeasureReference) -> Measure:
        """Retrieve the measure model object associated with the measure reference."""
        raise NotImplementedError

    @abstractmethod
    def get_entity_references(self) -> Sequence[EntityReference]:
        """Retrieve all entity references from the collection of semantic models."""
        raise NotImplementedError

    @abstractmethod
    def get_semantic_models_for_measure(self, measure_reference: MeasureReference) -> Sequence[SemanticModel]:
        """Retrieve a list of all semantic model model objects associated with the measure reference."""
        raise NotImplementedError

    @abstractmethod
    def get_agg_time_dimension_for_measure(self, measure_reference: MeasureReference) -> TimeDimensionReference:
        """Retrieves the aggregate time dimension that is associated with the measure reference."""

    @abstractmethod
    def get_entity_in_semantic_model(self, ref: SemanticModelElementReference) -> Optional[Entity]:
        """Retrieve the entity matching the element -> semantic model mapping, if any."""
        raise NotImplementedError

    @abstractmethod
    def get_by_reference(self, semantic_model_reference: SemanticModelReference) -> Optional[SemanticModel]:
        """Retrieve the semantic model object matching the input semantic model reference, if any."""
        raise NotImplementedError

    @property
    @abstractmethod
    def semantic_model_references(self) -> Sequence[SemanticModelReference]:
        """Return all SemanticModelReference objects associated with the semantic models in the collection."""
        raise NotImplementedError

    @abstractmethod
    def get_aggregation_time_dimensions_with_measures(
        self, semantic_model_reference: SemanticModelReference
    ) -> ElementGrouper[TimeDimensionReference, MeasureSpec]:
        """Return all aggregation time dimensions in the given semantic model with their associated measures."""
        raise NotImplementedError

    @abstractmethod
    def get_semantic_models_for_entity(self, entity_reference: EntityReference) -> Set[SemanticModel]:
        """Return all semantic models associated with an entity reference."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def entity_links_for_local_elements(semantic_model: SemanticModel) -> Sequence[EntityReference]:
        """Return the entity prefix that can be used to access dimensions defined in the semantic model."""
        raise NotImplementedError


class MetricAccessor(ABC):
    """Interface for accessing semantic information about a set of metric objects.

    This is primarily useful for restricting caller access to the subset of container methods and imports we want
    them to use. For example, the MetricLookup class might implement this protocol but also include some
    public methods for adding or removing metrics from the container, while this protocol only allows the
    caller to invoke the accessor methods which retrieve semantic information about the collected metrics.
    """

    @abstractmethod
    def element_specs_for_metrics(
        self,
        metric_references: Sequence[MetricReference],
        with_any_property: FrozenSet[LinkableElementProperties] = LinkableElementProperties.all_properties(),
        without_any_property: FrozenSet[LinkableElementProperties] = frozenset(),
    ) -> Sequence[LinkableInstanceSpec]:
        """Retrieve the matching set of linkable elements common to all metrics requested (intersection)."""
        raise NotImplementedError

    @abstractmethod
    def get_metrics(self, metric_references: Sequence[MetricReference]) -> Sequence[Metric]:
        """Retrieve the Metric model objects associated with the provided metric specs."""
        raise NotImplementedError

    @property
    @abstractmethod
    def metric_references(self) -> Sequence[MetricReference]:
        """Return the metric references."""
        raise NotImplementedError

    @abstractmethod
    def get_metric(self, metric_reference: MetricReference) -> Metric:  # noqa:D
        raise NotImplementedError

    @abstractmethod
    def measures_for_metric(
        self,
        metric_reference: MetricReference,
        column_association_resolver: ColumnAssociationResolver,
    ) -> Sequence[MetricInputMeasureSpec]:
        """Return the measure specs required to compute the metric."""
        raise NotImplementedError

    @abstractmethod
    def contains_cumulative_or_time_offset_metric(self, metric_references: Sequence[MetricReference]) -> bool:
        """Returns true if any of the specs correspond to a cumulative metric or a derived metric with time offset."""
        raise NotImplementedError

    @abstractmethod
    def metric_input_specs_for_metric(
        self,
        metric_reference: MetricReference,
        column_association_resolver: ColumnAssociationResolver,
    ) -> Sequence[MetricSpec]:
        """Returns the metric input specs required to compute the metric."""
        raise NotImplementedError



from __future__ import annotations

import logging
from typing import Dict, FrozenSet, List, Sequence

from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilterIntersection
from dbt_semantic_interfaces.implementations.metric import PydanticMetricTimeWindow
from dbt_semantic_interfaces.protocols import WhereFilter
from dbt_semantic_interfaces.protocols.metric import Metric, MetricType
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.references import MetricReference

from metricflow.errors.errors import DuplicateMetricError, MetricNotFoundError, NonExistentMeasureError
from metricflow.model.semantics.linkable_element_properties import LinkableElementProperties
from metricflow.model.semantics.linkable_spec_resolver import LinkableElementSet, ValidLinkableSpecResolver
from metricflow.model.semantics.semantic_model_join_evaluator import MAX_JOIN_HOPS
from metricflow.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow.protocols.semantics import MetricAccessor
from metricflow.specs.column_assoc import ColumnAssociationResolver
from metricflow.specs.specs import (
    LinkableInstanceSpec,
    MeasureSpec,
    MetricInputMeasureSpec,
    MetricSpec,
)
from metricflow.specs.where_filter_transform import WhereSpecFactory

logger = logging.getLogger(__name__)


class MetricLookup(MetricAccessor):  # noqa: D
    def __init__(  # noqa: D
        self, semantic_manifest: SemanticManifest, semantic_model_lookup: SemanticModelLookup
    ) -> None:
        self._semantic_manifest = semantic_manifest
        self._metrics: Dict[MetricReference, Metric] = {}
        self._semantic_model_lookup = semantic_model_lookup

        for metric in self._semantic_manifest.metrics:
            self.add_metric(metric)

        self._linkable_spec_resolver = ValidLinkableSpecResolver(
            semantic_manifest=self._semantic_manifest,
            semantic_model_lookup=semantic_model_lookup,
            max_entity_links=MAX_JOIN_HOPS,
        )

    def element_specs_for_metrics(
        self,
        metric_references: Sequence[MetricReference],
        with_any_property: FrozenSet[LinkableElementProperties] = LinkableElementProperties.all_properties(),
        without_any_property: FrozenSet[LinkableElementProperties] = frozenset(),
    ) -> Sequence[LinkableInstanceSpec]:
        """Dimensions common to all metrics requested (intersection)."""
        all_linkable_specs = self._linkable_spec_resolver.get_linkable_elements_for_metrics(
            metric_references=metric_references,
            with_any_of=with_any_property,
            without_any_of=without_any_property,
        ).as_spec_set

        return sorted(all_linkable_specs.as_tuple, key=lambda x: x.qualified_name)

    def linkable_set_for_metrics(
        self,
        metric_references: Sequence[MetricReference],
        with_any_property: FrozenSet[LinkableElementProperties] = LinkableElementProperties.all_properties(),
        without_any_property: FrozenSet[LinkableElementProperties] = frozenset(),
    ) -> LinkableElementSet:
        """Similar to element_specs_for_metrics(), but as a set with more context."""
        return self._linkable_spec_resolver.get_linkable_elements_for_metrics(
            metric_references=metric_references,
            with_any_of=with_any_property,
            without_any_of=without_any_property,
        )

    def get_metrics(self, metric_references: Sequence[MetricReference]) -> Sequence[Metric]:  # noqa: D
        res = []
        for metric_reference in metric_references:
            if metric_reference not in self._metrics:
                raise MetricNotFoundError(
                    f"Unable to find metric `{metric_reference}`. Perhaps it has not been registered"
                )
            res.append(self._metrics[metric_reference])

        return res

    @property
    def metric_references(self) -> Sequence[MetricReference]:  # noqa: D
        return list(self._metrics.keys())

    def get_metric(self, metric_reference: MetricReference) -> Metric:  # noqa:D
        if metric_reference not in self._metrics:
            raise MetricNotFoundError(f"Unable to find metric `{metric_reference}`. Perhaps it has not been registered")
        return self._metrics[metric_reference]

    def add_metric(self, metric: Metric) -> None:
        """Add metric, validating presence of required measures."""
        metric_reference = MetricReference(element_name=metric.name)
        if metric_reference in self._metrics:
            raise DuplicateMetricError(f"Metric `{metric.name}` has already been registered")
        for measure_reference in metric.measure_references:
            if measure_reference not in self._semantic_model_lookup.measure_references:
                raise NonExistentMeasureError(
                    f"Metric `{metric.name}` references measure `{measure_reference}` which has not been registered"
                )
        self._metrics[metric_reference] = metric

    def measures_for_metric(
        self,
        metric_reference: MetricReference,
        column_association_resolver: ColumnAssociationResolver,
    ) -> Sequence[MetricInputMeasureSpec]:
        """Return the measure specs required to compute the metric."""
        metric = self.get_metric(metric_reference)
        input_measure_specs: List[MetricInputMeasureSpec] = []

        for input_measure in metric.input_measures:
            measure_spec = MeasureSpec(
                element_name=input_measure.name,
                non_additive_dimension_spec=self._semantic_model_lookup.non_additive_dimension_specs_by_measure.get(
                    input_measure.measure_reference
                ),
            )
            spec = MetricInputMeasureSpec(
                measure_spec=measure_spec,
                constraint=WhereSpecFactory(
                    column_association_resolver=column_association_resolver,
                ).create_from_where_filter_intersection(input_measure.filter),
                alias=input_measure.alias,
            )
            input_measure_specs.append(spec)

        return tuple(input_measure_specs)

    def contains_cumulative_or_time_offset_metric(self, metric_references: Sequence[MetricReference]) -> bool:
        """Returns true if any of the specs correspond to a cumulative metric or a derived metric with time offset."""
        for metric_reference in metric_references:
            metric = self.get_metric(metric_reference)
            if metric.type == MetricType.CUMULATIVE:
                return True
            elif metric.type == MetricType.DERIVED:
                for input_metric in metric.type_params.metrics or []:
                    if input_metric.offset_window or input_metric.offset_to_grain:
                        return True
        return False

    def metric_input_specs_for_metric(
        self,
        metric_reference: MetricReference,
        column_association_resolver: ColumnAssociationResolver,
    ) -> Sequence[MetricSpec]:
        """Return the metric specs referenced by the metric. Current use case is for derived metrics."""
        metric = self.get_metric(metric_reference)
        input_metric_specs: List[MetricSpec] = []

        for input_metric in metric.input_metrics:
            original_metric_obj = self.get_metric(input_metric.as_reference)

            where_filters: List[WhereFilter] = []

            # This is the constraint parameter added to the input metric in the derived metric definition
            if input_metric.filter:
                where_filters.extend(input_metric.filter.where_filters)

            # This is the constraint parameter included in the original input metric definition
            if original_metric_obj.filter:
                where_filters.extend(original_metric_obj.filter.where_filters)

            spec = MetricSpec(
                element_name=input_metric.name,
                constraint=WhereSpecFactory(column_association_resolver).create_from_where_filter_intersection(
                    PydanticWhereFilterIntersection(where_filters=where_filters)
                ),
                alias=input_metric.alias,
                offset_window=PydanticMetricTimeWindow(
                    count=input_metric.offset_window.count,
                    granularity=input_metric.offset_window.granularity,
                )
                if input_metric.offset_window
                else None,
                offset_to_grain=input_metric.offset_to_grain,
            )
            input_metric_specs.append(spec)
        return tuple(input_metric_specs)



from __future__ import annotations

import pytest

from metricflow.api.metricflow_client import MetricFlowClient
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.protocols.sql_client import SqlClient


@pytest.fixture
def mf_client(
    create_source_tables: bool,
    sql_client: SqlClient,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> MetricFlowClient:
    """Fixture for MetricFlowClient."""
    return MetricFlowClient(
        sql_client=sql_client,
        semantic_manifest=simple_semantic_manifest_lookup.semantic_manifest,
    )



from __future__ import annotations

import datetime
import logging
from copy import copy
from typing import List, Optional, Sequence, Tuple

import jinja2
import pytest
from dateutil import parser
from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.implementations.elements.measure import PydanticMeasureAggregationParameters
from dbt_semantic_interfaces.test_utils import as_datetime
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.engine.metricflow_engine import MetricFlowEngine, MetricFlowQueryRequest
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import (
    DunderColumnAssociationResolver,
)
from metricflow.protocols.query_parameter import DimensionOrEntityQueryParameter
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs.query_param_implementations import DimensionOrEntityParameter, TimeDimensionParameter
from metricflow.sql.sql_exprs import (
    SqlCastToTimestampExpression,
    SqlColumnReference,
    SqlColumnReferenceExpression,
    SqlDateTruncExpression,
    SqlExtractExpression,
    SqlPercentileExpression,
    SqlPercentileExpressionArgument,
    SqlPercentileFunctionType,
    SqlStringExpression,
    SqlSubtractTimeIntervalExpression,
)
from metricflow.test.compare_df import assert_dataframes_equal
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.integration.configured_test_case import (
    CONFIGURED_INTEGRATION_TESTS_REPOSITORY,
    IntegrationTestModel,
    RequiredDwEngineFeatures,
)
from metricflow.test.time.configurable_time_source import (
    ConfigurableTimeSource,
)
from metricflow.time.date_part import DatePart

logger = logging.getLogger(__name__)


class CheckQueryHelpers:
    """Functions that can be used to help render check queries in integration tests."""

    def __init__(self, sql_client: SqlClient) -> None:  # noqa:D
        self._sql_client = sql_client

    def render_time_constraint(
        self,
        expr: str,
        start_time: str,
        stop_time: str,
    ) -> str:
        """Render an expression like "ds >='2020-01-01' AND ds < '2020-01-02'" for start_time = stop_time = '2020-01-01'."""
        start_expr = self.cast_to_ts(f"{start_time}")
        time_format = "%Y-%m-%d"
        stop_time_plus_one_day = (
            datetime.datetime.strptime(stop_time, time_format) + datetime.timedelta(days=1)
        ).strftime(time_format)
        stop_expr = self.cast_to_ts(f"{stop_time_plus_one_day}")
        return f"{expr} >= {start_expr} AND {expr} < {stop_expr}"

    def cast_expr_to_ts(self, expr: str) -> str:
        """Returns the expression as a new expression cast to the timestamp type, if applicable for the DB."""
        return f"CAST({expr} AS {self._sql_client.sql_query_plan_renderer.expr_renderer.timestamp_data_type})"

    def cast_to_ts(self, string_literal: str) -> str:
        """Similar to cast_expr_to_ts, but assumes the input string is to be converted into a string literal."""
        return self.cast_expr_to_ts(f"'{string_literal}'")

    def render_date_sub(
        self,
        table_alias: str,
        column_alias: str,
        count: int,
        granularity: TimeGranularity,
    ) -> str:
        """Renders a date subtract expression."""
        expr = SqlSubtractTimeIntervalExpression(
            arg=SqlColumnReferenceExpression(SqlColumnReference(table_alias, column_alias)),
            count=count,
            granularity=granularity,
        )
        return self._sql_client.sql_query_plan_renderer.expr_renderer.render_sql_expr(expr).sql

    def render_date_trunc(self, expr: str, granularity: TimeGranularity) -> str:
        """Return the DATE_TRUNC() call that can be used for converting the given expr to the granularity."""
        renderable_expr = SqlDateTruncExpression(
            time_granularity=granularity,
            arg=SqlCastToTimestampExpression(
                arg=SqlStringExpression(
                    sql_expr=expr,
                    requires_parenthesis=False,
                )
            ),
        )
        return self._sql_client.sql_query_plan_renderer.expr_renderer.render_sql_expr(renderable_expr).sql

    def render_extract(self, expr: str, date_part: DatePart) -> str:
        """Return the EXTRACT call that can be used for converting the given expr to the date_part."""
        renderable_expr = SqlExtractExpression(
            date_part=date_part,
            arg=SqlCastToTimestampExpression(
                arg=SqlStringExpression(
                    sql_expr=expr,
                    requires_parenthesis=False,
                )
            ),
        )
        return self._sql_client.sql_query_plan_renderer.expr_renderer.render_sql_expr(renderable_expr).sql

    def render_percentile_expr(
        self, expr: str, percentile: float, use_discrete_percentile: bool, use_approximate_percentile: bool
    ) -> str:
        """Return the percentile call that can be used for computing a percentile aggregation."""
        percentile_args = SqlPercentileExpressionArgument.from_aggregation_parameters(
            PydanticMeasureAggregationParameters(
                percentile=percentile,
                use_discrete_percentile=use_discrete_percentile,
                use_approximate_percentile=use_approximate_percentile,
            )
        )

        renderable_expr = SqlPercentileExpression(
            order_by_arg=SqlStringExpression(
                sql_expr=expr,
                requires_parenthesis=False,
            ),
            percentile_args=percentile_args,
        )
        return self._sql_client.sql_query_plan_renderer.expr_renderer.render_sql_expr(renderable_expr).sql

    @property
    def double_data_type_name(self) -> str:
        """Return the name of the double data type for the relevant SQL engine."""
        return self._sql_client.sql_query_plan_renderer.expr_renderer.double_data_type

    def render_dimension_template(self, dimension_name: str, entity_path: Sequence[str] = ()) -> str:
        """Renders a template that can be used to retrieve a dimension.

        For example:

         "{{ render_dimension_template('country_latest', entity_path=['listing']) }}"

         ->

         "{{ dimension('country_latest', entity_path=['listing'] }}"

         This is needed as the where_filter field in the definition files are rendered twice through Jinja - once
         by the test framework, and again by MF.
        """
        return f"{{{{ Dimension('{dimension_name}', entity_path={repr(entity_path)}) }}}}"

    def render_entity_template(self, entity_name: str, entity_path: Sequence[str] = ()) -> str:
        """Similar to render_dimension_template() but for entities."""
        return f"{{{{ Entity('{entity_name}', entity_path={repr(entity_path)}) }}}}"

    def render_time_dimension_template(
        self, time_dimension_name: str, time_granularity: str, entity_path: Sequence[str] = ()
    ) -> str:
        """Similar to render_dimension_template() but for time dimensions."""
        return (
            f"{{{{ TimeDimension('{time_dimension_name}', '{time_granularity}', entity_path={repr(entity_path)}) }}}}"
        )


def filter_not_supported_features(
    sql_client: SqlClient, required_features: Tuple[RequiredDwEngineFeatures, ...]
) -> Sequence[RequiredDwEngineFeatures]:
    """Given a list of required features, return a list of features not supported by the given SQLClient."""
    not_supported_features: List[RequiredDwEngineFeatures] = []
    for required_feature in required_features:
        if required_feature is RequiredDwEngineFeatures.CONTINUOUS_PERCENTILE_AGGREGATION:
            if not sql_client.sql_query_plan_renderer.expr_renderer.can_render_percentile_function(
                SqlPercentileFunctionType.CONTINUOUS
            ):
                not_supported_features.append(required_feature)
        elif required_feature is RequiredDwEngineFeatures.DISCRETE_PERCENTILE_AGGREGATION:
            if not sql_client.sql_query_plan_renderer.expr_renderer.can_render_percentile_function(
                SqlPercentileFunctionType.DISCRETE
            ):
                not_supported_features.append(required_feature)
        elif required_feature is RequiredDwEngineFeatures.APPROXIMATE_CONTINUOUS_PERCENTILE_AGGREGATION:
            if not sql_client.sql_query_plan_renderer.expr_renderer.can_render_percentile_function(
                SqlPercentileFunctionType.APPROXIMATE_CONTINUOUS
            ):
                not_supported_features.append(required_feature)
        elif required_feature is RequiredDwEngineFeatures.APPROXIMATE_DISCRETE_PERCENTILE_AGGREGATION:
            if not sql_client.sql_query_plan_renderer.expr_renderer.can_render_percentile_function(
                SqlPercentileFunctionType.APPROXIMATE_DISCRETE
            ):
                not_supported_features.append(required_feature)
        else:
            assert_values_exhausted(required_feature)
    return not_supported_features


@pytest.mark.parametrize(
    "name",
    CONFIGURED_INTEGRATION_TESTS_REPOSITORY.all_test_case_names,
    ids=lambda name: f"name={name}",
)
def test_case(
    name: str,
    mf_test_session_state: MetricFlowTestSessionState,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
    simple_semantic_manifest_lookup_non_ds: SemanticManifestLookup,
    unpartitioned_multi_hop_join_semantic_manifest_lookup: SemanticManifestLookup,
    multi_hop_join_semantic_manifest_lookup: SemanticManifestLookup,
    extended_date_semantic_manifest_lookup: SemanticManifestLookup,
    scd_semantic_manifest_lookup: SemanticManifestLookup,
    sql_client: SqlClient,
    create_source_tables: bool,
) -> None:
    """Runs all integration tests configured in the test case YAML directory."""
    case = CONFIGURED_INTEGRATION_TESTS_REPOSITORY.get_test_case(name)
    logger.info(f"Running integration test case: '{case.name}' from file '{case.file_path}'")

    missing_required_features = filter_not_supported_features(sql_client, case.required_features)
    if missing_required_features:
        pytest.skip(f"DW does not support {missing_required_features}")

    semantic_manifest_lookup: Optional[SemanticManifestLookup] = None
    if case.model is IntegrationTestModel.SIMPLE_MODEL:
        semantic_manifest_lookup = simple_semantic_manifest_lookup
    elif case.model is IntegrationTestModel.SIMPLE_MODEL_NON_DS:
        semantic_manifest_lookup = simple_semantic_manifest_lookup_non_ds
    elif case.model is IntegrationTestModel.UNPARTITIONED_MULTI_HOP_JOIN_MODEL:
        semantic_manifest_lookup = unpartitioned_multi_hop_join_semantic_manifest_lookup
    elif case.model is IntegrationTestModel.PARTITIONED_MULTI_HOP_JOIN_MODEL:
        semantic_manifest_lookup = multi_hop_join_semantic_manifest_lookup
    elif case.model is IntegrationTestModel.EXTENDED_DATE_MODEL:
        semantic_manifest_lookup = extended_date_semantic_manifest_lookup
    elif case.model is IntegrationTestModel.SCD_MODEL:
        semantic_manifest_lookup = scd_semantic_manifest_lookup
    else:
        assert_values_exhausted(case.model)

    assert semantic_manifest_lookup

    engine = MetricFlowEngine(
        semantic_manifest_lookup=semantic_manifest_lookup,
        sql_client=sql_client,
        column_association_resolver=DunderColumnAssociationResolver(semantic_manifest_lookup),
        time_source=ConfigurableTimeSource(as_datetime("2021-01-04")),
    )

    check_query_helpers = CheckQueryHelpers(sql_client)

    group_by: List[DimensionOrEntityQueryParameter] = []
    for group_by_kwargs in case.group_by_objs:
        kwargs = copy(group_by_kwargs)
        date_part = kwargs.get("date_part")
        grain = kwargs.get("grain")
        if date_part or grain:
            if date_part:
                kwargs["date_part"] = DatePart(date_part)
            if grain:
                kwargs["grain"] = TimeGranularity(grain)
            group_by.append(TimeDimensionParameter(**kwargs))
        else:
            group_by.append(DimensionOrEntityParameter(**kwargs))
    query_result = engine.query(
        MetricFlowQueryRequest.create_with_random_request_id(
            metric_names=case.metrics,
            group_by_names=case.group_bys,
            group_by=tuple(group_by),
            limit=case.limit,
            time_constraint_start=parser.parse(case.time_constraint[0]) if case.time_constraint else None,
            time_constraint_end=parser.parse(case.time_constraint[1]) if case.time_constraint else None,
            where_constraint=jinja2.Template(
                case.where_filter,
                undefined=jinja2.StrictUndefined,
            ).render(
                source_schema=mf_test_session_state.mf_source_schema,
                render_time_constraint=check_query_helpers.render_time_constraint,
                TimeGranularity=TimeGranularity,
                DatePart=DatePart,
                render_date_sub=check_query_helpers.render_date_sub,
                render_date_trunc=check_query_helpers.render_date_trunc,
                render_extract=check_query_helpers.render_extract,
                render_percentile_expr=check_query_helpers.render_percentile_expr,
                mf_time_spine_source=semantic_manifest_lookup.time_spine_source.spine_table.sql,
                double_data_type_name=check_query_helpers.double_data_type_name,
                render_dimension_template=check_query_helpers.render_dimension_template,
                render_entity_template=check_query_helpers.render_entity_template,
                render_time_dimension_template=check_query_helpers.render_time_dimension_template,
            )
            if case.where_filter
            else None,
            order_by_names=case.order_bys,
        )
    )

    actual = query_result.result_df

    expected = sql_client.query(
        jinja2.Template(
            case.check_query,
            undefined=jinja2.StrictUndefined,
        ).render(
            source_schema=mf_test_session_state.mf_source_schema,
            render_time_constraint=check_query_helpers.render_time_constraint,
            TimeGranularity=TimeGranularity,
            DatePart=DatePart,
            render_date_sub=check_query_helpers.render_date_sub,
            render_date_trunc=check_query_helpers.render_date_trunc,
            render_extract=check_query_helpers.render_extract,
            render_percentile_expr=check_query_helpers.render_percentile_expr,
            mf_time_spine_source=semantic_manifest_lookup.time_spine_source.spine_table.sql,
            double_data_type_name=check_query_helpers.double_data_type_name,
        )
    )
    # If we sort, it's effectively not checking the order whatever order that the output was would be overwritten.
    assert_dataframes_equal(actual, expected, sort_columns=not case.check_order, allow_empty=case.allow_empty)



from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Protocol

from pandas import DataFrame

from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_request.sql_request_attributes import SqlJsonTag


class SqlEngine(Enum):
    """Enumeration of supported SQL engines.

    Values are normalized engine names used for things like snapshot file locations.
    """

    BIGQUERY = "BigQuery"
    DUCKDB = "DuckDB"
    REDSHIFT = "Redshift"
    POSTGRES = "Postgres"
    SNOWFLAKE = "Snowflake"
    DATABRICKS = "Databricks"


class SqlClient(Protocol):
    """Base interface for SqlClient instances used inside MetricFlow.

    This provides the methods needed to execute SQL queries against the relevant Data Warehouse configuration.
    """

    @property
    @abstractmethod
    def sql_engine_type(self) -> SqlEngine:
        """Enumerated value representing the underlying SqlEngine for this SqlClient instance."""
        raise NotImplementedError

    @property
    @abstractmethod
    def sql_query_plan_renderer(self) -> SqlQueryPlanRenderer:
        """Dialect-specific SQL query plan renderer used for converting MetricFlow's query plan to executable SQL.

        This is bundled with the SqlClient partly as a convenenience for accessing a single instance of the renderer,
        and partly due to the close relationship between dialect and engine capabilities.
        """
        raise NotImplementedError

    @abstractmethod
    def query(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
        extra_tags: SqlJsonTag = SqlJsonTag(),
    ) -> DataFrame:
        """Base query method, upon execution will run a query that returns a pandas DataFrame."""
        raise NotImplementedError

    @abstractmethod
    def execute(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
        extra_tags: SqlJsonTag = SqlJsonTag(),
    ) -> None:
        """Base execute method."""
        raise NotImplementedError

    @abstractmethod
    def dry_run(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        """Base dry_run method."""
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:  # noqa: D
        """Close the connections / engines used by this client."""
        raise NotImplementedError

    @abstractmethod
    def render_bind_parameter_key(self, bind_parameter_key: str) -> str:
        """Wrap the bind parameter key with syntax accepted by engine."""
        raise NotImplementedError



from __future__ import annotations

import enum
import logging
import time

import pandas as pd
from dbt.adapters.base.impl import BaseAdapter
from dbt.exceptions import DbtDatabaseError
from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.pretty_print import pformat_big_objects

from metricflow.errors.errors import SqlBindParametersNotSupportedError
from metricflow.formatting import indent_log_line
from metricflow.protocols.sql_client import SqlEngine
from metricflow.random_id import random_id
from metricflow.sql.render.big_query import BigQuerySqlQueryPlanRenderer
from metricflow.sql.render.databricks import DatabricksSqlQueryPlanRenderer
from metricflow.sql.render.duckdb_renderer import DuckDbSqlQueryPlanRenderer
from metricflow.sql.render.postgres import PostgresSQLSqlQueryPlanRenderer
from metricflow.sql.render.redshift import RedshiftSqlQueryPlanRenderer
from metricflow.sql.render.snowflake import SnowflakeSqlQueryPlanRenderer
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_request.sql_request_attributes import SqlJsonTag, SqlRequestId, SqlRequestTagSet
from metricflow.sql_request.sql_statement_metadata import CombinedSqlTags, SqlStatementCommentMetadata

logger = logging.getLogger(__name__)


# Discovered via trial and error from the original, and now defunct DatabricksSqlClient implementation
DATABRICKS_SQL_WAREHOUSE_EXPLAIN_PLAN_ERROR_KEY = "Error occurred during query planning"
DATABRICKS_CLUSTER_EXPLAIN_PLAN_ERROR_KEY = "org.apache.spark.sql.AnalysisException"


class SupportedAdapterTypes(enum.Enum):
    """Enumeration of supported dbt adapter types."""

    DATABRICKS = "databricks"
    POSTGRES = "postgres"
    SNOWFLAKE = "snowflake"
    REDSHIFT = "redshift"
    BIGQUERY = "bigquery"
    DUCKDB = "duckdb"

    @property
    def sql_engine_type(self) -> SqlEngine:
        """Return the SqlEngine corresponding to the supported adapter type."""
        if self is SupportedAdapterTypes.BIGQUERY:
            return SqlEngine.BIGQUERY
        elif self is SupportedAdapterTypes.DATABRICKS:
            return SqlEngine.DATABRICKS
        elif self is SupportedAdapterTypes.POSTGRES:
            return SqlEngine.POSTGRES
        elif self is SupportedAdapterTypes.REDSHIFT:
            return SqlEngine.REDSHIFT
        elif self is SupportedAdapterTypes.SNOWFLAKE:
            return SqlEngine.SNOWFLAKE
        elif self is SupportedAdapterTypes.DUCKDB:
            return SqlEngine.DUCKDB
        else:
            assert_values_exhausted(self)

    @property
    def sql_query_plan_renderer(self) -> SqlQueryPlanRenderer:
        """Return the SqlQueryPlanRenderer corresponding to the supported adapter type."""
        if self is SupportedAdapterTypes.BIGQUERY:
            return BigQuerySqlQueryPlanRenderer()
        elif self is SupportedAdapterTypes.DATABRICKS:
            return DatabricksSqlQueryPlanRenderer()
        elif self is SupportedAdapterTypes.POSTGRES:
            return PostgresSQLSqlQueryPlanRenderer()
        elif self is SupportedAdapterTypes.REDSHIFT:
            return RedshiftSqlQueryPlanRenderer()
        elif self is SupportedAdapterTypes.SNOWFLAKE:
            return SnowflakeSqlQueryPlanRenderer()
        elif self is SupportedAdapterTypes.DUCKDB:
            return DuckDbSqlQueryPlanRenderer()
        else:
            assert_values_exhausted(self)


class AdapterBackedSqlClient:
    """SqlClient implementation which delegates database operations to a dbt BaseAdapter instance.

    This is a generic wrpaper class meant to cover all of our logging, querying, and internal configuration
    needs while delegating all connection state management and warehouse communication work to an underlying
    dbt adapter instance. This relies on BaseAdapter, rather than SQLAdapter, because BigQuery is an instance
    of the more generic BaseAdapter class.
    """

    def __init__(self, adapter: BaseAdapter):
        """Initializer sourced from a BaseAdapter instance.

        The dbt BaseAdapter should already be fully initialized, including all credential verification, and
        ready for use for establishing connections and issuing queries.
        """
        self._adapter = adapter
        try:
            adapter_type = SupportedAdapterTypes(self._adapter.type())
        except ValueError as e:
            raise ValueError(
                f"Adapter type {self._adapter.type()} is not supported. Must be one "
                f"of {[item.value for item in SupportedAdapterTypes]}."
            ) from e

        self._sql_engine_type = adapter_type.sql_engine_type
        self._sql_query_plan_renderer = adapter_type.sql_query_plan_renderer
        logger.info(f"Initialized AdapterBackedSqlClient with dbt adapter type `{adapter_type.value}`")

    @property
    def sql_engine_type(self) -> SqlEngine:
        """An enumerated value representing the underlying SqlEngine supported by the dbt adapter for this instance."""
        return self._sql_engine_type

    @property
    def sql_query_plan_renderer(self) -> SqlQueryPlanRenderer:
        """Dialect-specific SQL query plan renderer used for converting MetricFlow's query plan to executable SQL."""
        return self._sql_query_plan_renderer

    def query(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
        extra_tags: SqlJsonTag = SqlJsonTag(),
    ) -> pd.DataFrame:
        """Query statement; result expected to be data which will be returned as a DataFrame.

        Args:
            stmt: The SQL query statement to run. This should produce output via a SELECT
            sql_bind_parameters: The parameter replacement mapping for filling in
                concrete values for SQL query parameters.
            extra_tags: An object containing JSON serialized tags meant for annotating queries.
        """
        start = time.time()
        request_id = SqlRequestId(f"mf_rid__{random_id()}")
        combined_tags = AdapterBackedSqlClient._consolidate_tags(json_tags=extra_tags, request_id=request_id)
        statement = SqlStatementCommentMetadata.add_tag_metadata_as_comment(
            sql_statement=stmt, combined_tags=combined_tags
        )
        if sql_bind_parameters.param_dict:
            raise SqlBindParametersNotSupportedError(
                f"Invalid execute statement - we do not support queries with bind parameters through dbt adapters! "
                f"Bind params: {sql_bind_parameters.param_dict}"
            )
        logger.info(AdapterBackedSqlClient._format_run_query_log_message(statement, sql_bind_parameters))
        with self._adapter.connection_named(f"MetricFlow_request_{request_id}"):
            # returns a Tuple[AdapterResponse, agate.Table] but the decorator converts it to Any
            result = self._adapter.execute(sql=statement, auto_begin=True, fetch=True)
            logger.info(f"Query returned from dbt Adapter with response {result[0]}")

        agate_data = result[1]
        df = pd.DataFrame([row.values() for row in agate_data.rows], columns=agate_data.column_names)
        stop = time.time()
        logger.info(f"Finished running the query in {stop - start:.2f}s with {df.shape[0]} row(s) returned")
        return df

    def execute(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
        extra_tags: SqlJsonTag = SqlJsonTag(),
    ) -> None:
        """Execute a SQL statement. No result will be returned.

        Args:
            stmt: The SQL query statement to run. This should not produce output.
            sql_bind_parameters: The parameter replacement mapping for filling in
                concrete values for SQL query parameters.
            extra_tags: An object containing JSON serialized tags meant for annotating queries.
        """
        if sql_bind_parameters.param_dict:
            raise SqlBindParametersNotSupportedError(
                f"Invalid execute statement - we do not support execute commands with bind parameters through dbt "
                f"adapters! Bind params: {SqlBindParameters.param_dict}"
            )
        start = time.time()
        request_id = SqlRequestId(f"mf_rid__{random_id()}")
        combined_tags = AdapterBackedSqlClient._consolidate_tags(json_tags=extra_tags, request_id=request_id)
        statement = SqlStatementCommentMetadata.add_tag_metadata_as_comment(
            sql_statement=stmt, combined_tags=combined_tags
        )
        logger.info(AdapterBackedSqlClient._format_run_query_log_message(statement, sql_bind_parameters))
        with self._adapter.connection_named(f"MetricFlow_request_{request_id}"):
            result = self._adapter.execute(statement, auto_begin=True, fetch=False)
            # Calls to execute often involve some amount of DDL so we commit here
            self._adapter.commit_if_has_connection()
            logger.info(f"Query executed via dbt Adapter with response {result[0]}")
        stop = time.time()
        logger.info(f"Finished running the query in {stop - start:.2f}s")
        return None

    def dry_run(
        self,
        stmt: str,
        sql_bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        """Dry run statement; checks that the 'stmt' is queryable. Returns None on success.

        Raises an exception if the 'stmt' isn't queryable.

        Args:
            stmt: The SQL query statement to dry run.
            sql_bind_parameters: The parameter replacement mapping for filling in
                concrete values for SQL query parameters.
        """
        start = time.time()
        logger.info(
            f"Running dry_run of:"
            f"\n\n{indent_log_line(stmt)}\n"
            + (f"\nwith parameters: {dict(sql_bind_parameters.param_dict)}" if sql_bind_parameters.param_dict else "")
        )
        request_id = SqlRequestId(f"mf_rid__{random_id()}")
        connection_name = f"MetricFlow_dry_run_request_{request_id}"
        # TODO - consolidate to self._adapter.validate_sql() when all implementations will work from within MetricFlow
        if self.sql_engine_type is SqlEngine.BIGQUERY:
            with self._adapter.connection_named(connection_name):
                self._adapter.validate_sql(stmt)
        else:
            is_databricks = self.sql_engine_type is SqlEngine.DATABRICKS
            with self._adapter.connection_named(connection_name):
                results = self._adapter.execute(f"EXPLAIN {stmt}", auto_begin=True, fetch=is_databricks)

            if is_databricks:
                # We have to parse the output results from adapter.execute, but only in Databricks. We should probably use
                # a subclass for this and override things properly, but there's hopefully an upstream fix in our near
                # future so let's just put this ugly bit in for now.
                plan_output_str = "\n".join([",".join(row.values()) for row in results[1].rows])
                has_error = (
                    plan_output_str.find(DATABRICKS_CLUSTER_EXPLAIN_PLAN_ERROR_KEY) != -1
                    or plan_output_str.find(DATABRICKS_SQL_WAREHOUSE_EXPLAIN_PLAN_ERROR_KEY) != -1
                )
                if has_error:
                    raise DbtDatabaseError(f"Encountered error in Databricks dry run. Full output: {plan_output_str}")

        stop = time.time()
        logger.info(f"Finished running the dry_run in {stop - start:.2f}s")
        return

    def close(self) -> None:  # noqa: D
        self._adapter.cancel_open_connections()

    def render_bind_parameter_key(self, bind_parameter_key: str) -> str:
        """Wrap execution parameter key with syntax accepted by engine."""
        raise SqlBindParametersNotSupportedError(
            "We do not support queries with bind parameters through dbt adapters, so we do not have rendering enabled!"
        )

    @staticmethod
    def _format_run_query_log_message(statement: str, sql_bind_parameters: SqlBindParameters) -> str:
        """Helper for creating nicely formatted query logging."""
        message = f"Running query:\n\n{indent_log_line(statement)}"
        if len(sql_bind_parameters.param_dict) > 0:
            message += (
                f"\n"
                f"\n"
                f"with parameters:\n"
                f"\n"
                f"{indent_log_line(pformat_big_objects(sql_bind_parameters.param_dict))}"
            )
        return message

    @staticmethod
    def _consolidate_tags(json_tags: SqlJsonTag, request_id: SqlRequestId) -> CombinedSqlTags:
        """Consolidates json tags and request ID into a single set of tags."""
        return CombinedSqlTags(
            system_tags=SqlRequestTagSet().add_request_id(request_id=request_id),
            extra_tag=json_tags,
        )



from __future__ import annotations

import shutil
import textwrap
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator

import pytest
from dbt_semantic_interfaces.parsing.dir_to_model import (
    parse_yaml_files_to_validation_ready_semantic_manifest,
)
from dbt_semantic_interfaces.parsing.objects import YamlConfigFile
from dbt_semantic_interfaces.test_utils import base_semantic_manifest_file

from metricflow.cli.main import (
    dimension_values,
    dimensions,
    entities,
    health_checks,
    metrics,
    query,
    tutorial,
    validate_configs,
)
from metricflow.protocols.sql_client import SqlEngine
from metricflow.test.fixtures.cli_fixtures import MetricFlowCliRunner
from metricflow.test.model.example_project_configuration import EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE


def test_query(capsys: pytest.CaptureFixture, cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Disabling capsys to resolve error "ValueError: I/O operation on closed file". Better solution TBD.
    with capsys.disabled():
        resp = cli_runner.run(query, args=["--metrics", "bookings", "--group-by", "metric_time"])
    # case insensitive matches are needed for snowflake due to the capitalization thing
    engine_is_snowflake = cli_runner.cli_context.sql_client.sql_engine_type is SqlEngine.SNOWFLAKE
    assert "bookings" in resp.output or ("bookings" in resp.output.lower() and engine_is_snowflake)
    assert resp.exit_code == 0


def test_list_dimensions(capsys: pytest.CaptureFixture, cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Disabling capsys to resolve error "ValueError: I/O operation on closed file". Better solution TBD.
    with capsys.disabled():
        resp = cli_runner.run(dimensions, args=["--metrics", "bookings"])

    assert "ds" in resp.output
    assert resp.exit_code == 0


def test_list_metrics(capsys: pytest.CaptureFixture, cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Disabling capsys to resolve error "ValueError: I/O operation on closed file". Better solution TBD.
    with capsys.disabled():
        resp = cli_runner.run(metrics)

    assert "bookings_per_listing: listing__capacity_latest" in resp.output
    assert resp.exit_code == 0


def test_get_dimension_values(capsys: pytest.CaptureFixture, cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Disabling capsys to resolve error "ValueError: I/O operation on closed file". Better solution TBD.
    with capsys.disabled():
        resp = cli_runner.run(dimension_values, args=["--metrics", "bookings", "--dimension", "booking__is_instant"])

    actual_output_lines = sorted(resp.output.split("\n"))
    assert ["", "• False", "• True"] == actual_output_lines
    assert resp.exit_code == 0


@contextmanager
def create_directory(directory_path: str) -> Iterator[None]:
    """Creates a temporary directory (errors if it exists) and removes it."""
    path = Path(directory_path)
    path.mkdir(parents=True, exist_ok=True)
    yield
    shutil.rmtree(path)


def test_validate_configs(capsys: pytest.CaptureFixture, cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    yaml_contents = textwrap.dedent(
        """\
        semantic_model:
          name: bad_semantic_model
          node_relation:
            schema_name: some_schema
            alias: some_table
          defaults:
            agg_time_dimension: ds
          dimensions:
            - name: country
              type: categorical
        """
    )
    bad_semantic_model = YamlConfigFile(filepath="inline_for_test", contents=yaml_contents)
    # JSON-stored manifests from dbt are not transformed, so we run this test on that style of output
    manifest = parse_yaml_files_to_validation_ready_semantic_manifest(
        [EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, base_semantic_manifest_file(), bad_semantic_model],
        apply_transformations=False,
    ).semantic_manifest

    target_directory = Path().absolute() / "target"
    with create_directory(target_directory.as_posix()):
        manifest_file = target_directory / "semantic_manifest.json"
        manifest_file.write_text(manifest.json())

        # Disabling capsys to resolve error "ValueError: I/O operation on closed file". Better solution TBD.
        with capsys.disabled():
            resp = cli_runner.run(validate_configs)

        assert "ERROR" in resp.output
        assert resp.exit_code == 0


def test_health_checks(capsys: pytest.CaptureFixture, cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Disabling capsys to resolve error "ValueError: I/O operation on closed file". Better solution TBD.
    with capsys.disabled():
        resp = cli_runner.run(health_checks)

    assert "SELECT 1: Success!" in resp.output
    assert resp.exit_code == 0


def test_tutorial_message(capsys: pytest.CaptureFixture, cli_runner: MetricFlowCliRunner) -> None:
    """Tests the message output of the tutorial.

    The tutorial now essentially compiles a semantic manifest and then asks the user to run dbt seed,
    so from an end user perspective it's little more than the output with -m.

    The tutorial currently requires execution from a dbt project path. Rather than go all the way on testing the
    tutorial given the path and dbt project requirements, we simply check the message output. When we allow for
    project path overrides it might warrant a more complete test of the semantic manifest building steps in the
    tutorial flow.
    """
    # Disabling capsys to resolve error "ValueError: I/O operation on closed file". Better solution TBD.
    with capsys.disabled():
        resp = cli_runner.run(tutorial, args=["-m"])
    assert "Please run the following steps" in resp.output
    assert "dbt seed" in resp.output


def test_list_entities(capsys: pytest.CaptureFixture, cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Disabling capsys to resolve error "ValueError: I/O operation on closed file". Better solution TBD.
    with capsys.disabled():
        resp = cli_runner.run(entities, args=["--metrics", "bookings"])

    assert "guest" in resp.output
    assert "host" in resp.output



from __future__ import annotations

import datetime as dt
import logging
import pathlib
import traceback
from functools import update_wrapper, wraps
from typing import Any, Callable, List, Optional

import click
from dateutil.parser import parse

import metricflow.cli.custom_click_types as click_custom
from metricflow.cli.cli_context import CLIContext

logger = logging.getLogger(__name__)


# Click Options
def query_options(function: Callable) -> Callable:
    """Common options for a query."""
    function = click.option(
        "--order",
        type=click_custom.SequenceParamType(),
        help='Metrics or group bys to order by ("-" prefix for DESC). For example: --order -ds or --order ds,-revenue',
        required=False,
    )(function)
    function = click.option(
        "--limit",
        type=str,
        help="Limit the number of rows out using an int or leave blank for no limit. For example: --limit 100",
        callback=lambda ctx, param, value: validate_limit(value),
    )(function)
    function = click.option(
        "--where",
        type=str,
        default=None,
        help='SQL-like where statement provided as a string. For example: --where "revenue > 100"',
    )(function)
    function = start_end_time_options(function)
    function = click.option(
        "--group-by",
        type=click_custom.SequenceParamType(),
        default="",
        help="Dimensions and/or entities to group by: syntax is --group-by ds or for multiple group bys --group-by ds,org",
    )(function)
    function = click.option(
        "--metrics",
        type=click_custom.SequenceParamType(min_length=1),
        default="",
        help="Metrics to query for: syntax is --metrics bookings or for multiple metrics --metrics bookings,messages",
    )(function)
    return function


def start_end_time_options(function: Callable) -> Callable:
    """Options for start_time and end_time."""
    function = click.option(
        "--start-time",
        type=str,
        default=None,
        help="Optional iso8601 timestamp to constraint the start time of the data (inclusive)",
        callback=lambda ctx, param, value: convert_to_datetime(value),
    )(function)

    function = click.option(
        "--end-time",
        type=str,
        default=None,
        help="Optional iso8601 timestamp to constraint the end time of the data (inclusive)",
        callback=lambda ctx, param, value: convert_to_datetime(value),
    )(function)
    return function


# Parsers/Validators
def convert_to_datetime(datetime_str: Optional[str]) -> Optional[dt.datetime]:
    """Callback to convert string to datetime given as an iso8601 timestamp."""
    if datetime_str is None:
        return None

    try:
        return parse(datetime_str)
    except Exception:
        raise click.BadParameter("must be valid iso8601 timestamp")


def parse_comma_separated_inputs(value: Optional[str]) -> Optional[List[str]]:  # noqa: D
    # If comma exist, explode this into a list and return
    if value is None:
        return None
    if "," in value:
        return [i.strip() for i in value.split(",")]

    # Return a list of the single value
    return [value]


def validate_limit(limit: Optional[str]) -> Optional[int]:
    """Validates and transform limit input."""
    if limit and not limit.isnumeric():
        raise click.BadParameter("limit must be an int. For no limit, do not pass this argument")
    return int(limit) if limit else None


# Misc
def exception_handler(func: Callable[..., Any]) -> Callable[..., Any]:  # type: ignore[misc]
    """Decorator to handle exceptions."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:  # type: ignore[misc]
        try:
            func(*args, **kwargs)
        except Exception as e:
            # This will log to the file handlers registered in the root.
            logging.exception("Got an exception in the exception handler.")
            # Checks if CLIContext has verbose flag set

            if isinstance(args[0], CLIContext):
                cli_context: CLIContext = args[0]
                click.echo(f"\nERROR: {str(e)}\nLog file: {cli_context.log_file_path}")
            else:
                if not isinstance(args[0], CLIContext):
                    logger.error(
                        f"Missing {CLIContext.__name__} as the first argument to the function "
                        f"{getattr(func, '__name__', repr(func))}"
                    )
                click.echo(f"\nERROR: {str(e)}")
            if args and hasattr(args[0], "verbose") and args[0].verbose is True:
                click.echo(traceback.format_exc())

            exit(1)

    return wrapper


def dbt_project_file_exists() -> bool:
    """Check that the cwd is a dbt project root. Currently done by checking for existence of dbt_project.yml."""
    return pathlib.Path("dbt_project.yml").exists()


def error_if_not_in_dbt_project(func: Callable) -> Callable:
    """Decorator to output an error message and exit if caller is not in a root directory of a dbt project."""

    @click.pass_context
    def new_func(ctx: click.core.Context, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc]
        if not dbt_project_file_exists():
            click.echo(
                "❌ Unable to locate 'dbt_project.yml' in the current directory\n"
                "In order to run the MetricFlow CLI, you must be running in the root directory of a working dbt project.\n"
                "Please check out `https://docs.getdbt.com/reference/commands/init` if you want to get started on building a dbt project."
            )
            exit(1)
        return ctx.invoke(func, *args, **kwargs)

    return update_wrapper(new_func, func)



from __future__ import annotations

import logging

from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.naming.linkable_spec_name import DUNDER, StructuredLinkableSpecName
from metricflow.specs.column_assoc import (
    ColumnAssociation,
    ColumnAssociationResolver,
    SingleColumnCorrelationKey,
)
from metricflow.specs.specs import (
    DimensionSpec,
    EntitySpec,
    InstanceSpec,
    InstanceSpecVisitor,
    MeasureSpec,
    MetadataSpec,
    MetricSpec,
    TimeDimensionSpec,
)

logger = logging.getLogger(__name__)


class DunderColumnAssociationResolverVisitor(InstanceSpecVisitor[ColumnAssociation]):
    """Visitor helper class for DefaultColumnAssociationResolver2."""

    def __init__(self, semantic_manifest_lookup: SemanticManifestLookup) -> None:  # noqa: D
        self._semantic_manifest_lookup = semantic_manifest_lookup

    def visit_metric_spec(self, metric_spec: MetricSpec) -> ColumnAssociation:  # noqa: D
        return ColumnAssociation(
            column_name=metric_spec.element_name,
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def visit_measure_spec(self, measure_spec: MeasureSpec) -> ColumnAssociation:  # noqa: D
        return ColumnAssociation(
            column_name=measure_spec.element_name,
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def visit_dimension_spec(self, dimension_spec: DimensionSpec) -> ColumnAssociation:  # noqa: D
        return ColumnAssociation(
            column_name=StructuredLinkableSpecName(
                entity_link_names=tuple(x.element_name for x in dimension_spec.entity_links),
                element_name=dimension_spec.element_name,
            ).qualified_name,
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def visit_time_dimension_spec(self, time_dimension_spec: TimeDimensionSpec) -> ColumnAssociation:  # noqa: D
        column_name = StructuredLinkableSpecName(
            entity_link_names=tuple(x.element_name for x in time_dimension_spec.entity_links),
            element_name=time_dimension_spec.element_name,
            time_granularity=time_dimension_spec.time_granularity,
            date_part=time_dimension_spec.date_part,
        ).qualified_name

        return ColumnAssociation(
            column_name=column_name
            + (
                f"{DUNDER}{time_dimension_spec.aggregation_state.value.lower()}"
                if time_dimension_spec.aggregation_state
                else ""
            ),
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def visit_entity_spec(self, entity_spec: EntitySpec) -> ColumnAssociation:  # noqa: D
        return ColumnAssociation(
            column_name=StructuredLinkableSpecName(
                entity_link_names=tuple(x.element_name for x in entity_spec.entity_links),
                element_name=entity_spec.element_name,
            ).qualified_name,
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )

    def visit_metadata_spec(self, metadata_spec: MetadataSpec) -> ColumnAssociation:  # noqa: D
        return ColumnAssociation(
            column_name=metadata_spec.element_name,
            single_column_correlation_key=SingleColumnCorrelationKey(),
        )


class DunderColumnAssociationResolver(ColumnAssociationResolver):
    """Uses a double underscore to map specs to column names.

    For example:

    DimensionSpec(element_name='country', entity_links=['listing'])

    ->

    listing__country
    """

    def __init__(self, semantic_manifest_lookup: SemanticManifestLookup) -> None:  # noqa: D
        self._visitor_helper = DunderColumnAssociationResolverVisitor(semantic_manifest_lookup)

    def resolve_spec(self, spec: InstanceSpec) -> ColumnAssociation:  # noqa: D
        return spec.accept(self._visitor_helper)



from __future__ import annotations

import datetime as dt
import logging
import pathlib
import signal
import sys
import tempfile
import textwrap
import time
import warnings
from importlib.metadata import version as pkg_version
from typing import Callable, List, Optional

import click
import jinja2
import pandas as pd
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.validations.semantic_manifest_validator import SemanticManifestValidator
from dbt_semantic_interfaces.validations.validator_helpers import SemanticManifestValidationResults
from halo import Halo
from packaging.version import parse
from update_checker import UpdateChecker

import metricflow.cli.custom_click_types as click_custom
from metricflow.cli import PACKAGE_NAME
from metricflow.cli.cli_context import CLIContext
from metricflow.cli.constants import DEFAULT_RESULT_DECIMAL_PLACES, MAX_LIST_OBJECT_ELEMENTS
from metricflow.cli.dbt_connectors.dbt_config_accessor import dbtArtifacts
from metricflow.cli.tutorial import (
    dbtMetricFlowTutorialHelper,
)
from metricflow.cli.utils import (
    dbt_project_file_exists,
    error_if_not_in_dbt_project,
    exception_handler,
    query_options,
    start_end_time_options,
)
from metricflow.dag.dag_visualization import display_dag_as_svg
from metricflow.dataflow.dataflow_plan_to_text import dataflow_plan_as_text
from metricflow.engine.metricflow_engine import MetricFlowExplainResult, MetricFlowQueryRequest, MetricFlowQueryResult
from metricflow.model.data_warehouse_model_validator import DataWarehouseModelValidator
from metricflow.telemetry.models import TelemetryLevel
from metricflow.telemetry.reporter import TelemetryReporter, log_call

logger = logging.getLogger(__name__)

pass_config = click.make_pass_decorator(CLIContext, ensure=True)
_telemetry_reporter = TelemetryReporter(report_levels_higher_or_equal_to=TelemetryLevel.USAGE)
_telemetry_reporter.add_python_log_handler()
_telemetry_reporter.add_rudderstack_handler()


@click.group()
@click.option("-v", "--verbose", is_flag=True)
@click.version_option()
@error_if_not_in_dbt_project
@pass_config
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def cli(cfg: CLIContext, verbose: bool) -> None:  # noqa: D
    # Some HTTP logging callback somewhere is failing to close its SSL connections correctly.
    # For now, filter those warnings so they don't pop up in CLI stderr
    # note - this should be addressed as adapter connection issues might produce these as well
    warnings.filterwarnings("ignore", category=ResourceWarning, message="unclosed.*<ssl.SSLSocket.*>")

    cfg.verbose = verbose

    checker = UpdateChecker()
    result = checker.check(PACKAGE_NAME, pkg_version(PACKAGE_NAME))
    # result is None when an update was not found or a failure occurred
    if result:
        click.secho(
            "‼️ Warning: A new version of the MetricFlow CLI is available.",
            bold=True,
            fg="red",
        )

        click.echo(
            f"💡 Please update to version {result.available_version}, released {result.release_date} by running:\n"
            f"\t$ pip install --upgrade {PACKAGE_NAME}\n",
        )

    # Cancel queries submitted to the DW if the user precess CTRL + c / process is terminated.
    # Note: docs unclear on the type for the 'frame' argument.
    def exit_signal_handler(signal_type: int, frame) -> None:  # type: ignore
        if signal_type == signal.SIGINT:
            click.echo("Got SIGINT")
        elif signal_type == signal.SIGTERM:
            click.echo("Got SIGTERM")
        else:
            # Shouldn't happen since this should ony be registered for SIGINT / SIGTERM.
            click.echo(f"Got unhandled signal {signal_type}")
            return

        try:
            # Note: we may wish to add support for canceling all queries if zombie queries are a problem
            logger.info("Closing client connections")
            cfg.sql_client.close()
        finally:
            sys.exit(-1)

    signal.signal(signal.SIGINT, exit_signal_handler)
    signal.signal(signal.SIGTERM, exit_signal_handler)


@cli.command()
@click.option("-m", "--msg", is_flag=True, help="Output the final steps dialogue")
# @click.option("--skip-dw", is_flag=True, help="Skip the data warehouse health checks") # TODO: re-enable this
@click.option("--clean", is_flag=True, help="Remove sample model files.")
@pass_config
@click.pass_context
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def tutorial(ctx: click.core.Context, cfg: CLIContext, msg: bool, clean: bool) -> None:
    """Run user through a tutorial."""
    help_msg = textwrap.dedent(
        """\
        🤓 Please run the following steps,
            1.  Verify that your adapter credentials are correct in `profiles.yml`
            2.  Run `dbt seed`, check to see that the steps related to countries, transactions, customers are passing.
            3.  Try validating your data model: `mf validate-configs`
            4.  Check out your metrics: `mf list metrics`
            5.  Check out dimensions for your metric `mf list dimensions --metrics transactions`
            6.  Query your first metric: `mf query --metrics transactions --group-bys metric_time --order metric_time`
            7.  Show the SQL MetricFlow generates:
                `mf query --metrics transactions --group-bys metric_time --order metric_time --explain`
            8.  Visualize the plan:
                `mf query --metrics transactions --group-bys metric_time --order metric_time --explain --display-plans`
                * This only works if you have graphviz installed - see README.
            9.  Add another dimension:
                `mf query --metrics transactions --group-bys metric_time,customer__customer_country --order metric_time`
            10.  Add a coarser time granularity:
                `mf query --metrics transactions --group-bys metric_time__week --order metric_time__week`
            11. Try a more complicated query: mf query --metrics transactions,transaction_usd_na --group-bys metric_time,is_large --order metric_time --start-time 2022-03-20 --end-time 2022-04-01.
            12. For more ways to interact with the sample models, go to
                ‘https://docs.transform.co/docs/metricflow/metricflow-tutorial’.
            13. When you're done with the tutorial, run mf tutorial --clean to delete sample models and seeds.
        """
    )

    if msg:
        click.echo(help_msg)
        exit()

    if not dbt_project_file_exists():
        click.echo(
            "Unable to detect dbt project. Please ensure that your current working directory is at the root of the dbt project."
        )
        exit()

    # TODO: Health checks

    # Load the metadata from dbt project
    try:
        dbt_project_metadata = cfg.dbt_project_metadata
        dbt_paths = dbt_project_metadata.dbt_paths
        model_path = pathlib.Path(dbt_paths.model_paths[0]) / "sample_model"
        seed_path = pathlib.Path(dbt_paths.seed_paths[0]) / "sample_seed"
        manifest_path = pathlib.Path(dbt_paths.target_path) / "semantic_manifest.json"
    except Exception as e:
        click.echo(f"Unable to parse path metadata from dbt project.\nERROR: {str(e)}")
        exit(1)

    # Remove sample files from dbt project
    if clean:
        click.confirm("Would you like to remove all the sample files?", abort=True)
        spinner = Halo(text="Removing sample files...", spinner="dots")
        spinner.start()
        try:
            dbtMetricFlowTutorialHelper.remove_sample_files(model_path=model_path, seed_path=seed_path)
            spinner.succeed("🗑️ Sample files has been removed.")
            exit()
        except Exception as e:
            spinner.fail(f"❌ Unable to remove sample files.\nERROR: {str(e)}")
            exit(1)

    click.echo(
        textwrap.dedent(
            f"""\
        To begin building and querying metrics, you must define semantic models and
        metric configuration files in your dbt project. dbt will use these files to generate a
        semantic manifest artifact, which MetricFlow will use to create a semantic graph for querying.
        As part of this tutorial, we will generate the following files to help you get started:

        📜 model files -> {model_path.absolute().as_posix()}
        🌱 seed files -> {seed_path.absolute().as_posix()}
        ✅ semantic manifest json file -> {manifest_path.absolute().as_posix()}
        """
        )
    )
    click.confirm("Continue and generate the files?", abort=True)

    # Generate sample files into dbt project
    if dbtMetricFlowTutorialHelper.check_if_path_exists([model_path, seed_path]):
        click.confirm("There are existing files in the paths above, would you like to overwrite them?", abort=True)
        dbtMetricFlowTutorialHelper.remove_sample_files(model_path=model_path, seed_path=seed_path)

    spinner = Halo(text="Generating sample files...", spinner="dots")
    spinner.start()
    dbtMetricFlowTutorialHelper.generate_model_files(model_path=model_path, profile_schema=dbt_project_metadata.schema)
    dbtMetricFlowTutorialHelper.generate_seed_files(seed_path=seed_path)
    dbtMetricFlowTutorialHelper.generate_semantic_manifest_file(manifest_path=manifest_path)

    spinner.succeed("📜 Sample files has been generated.")

    click.echo(help_msg)
    click.echo("💡 Run `mf tutorial --msg` to see this message again without executing everything else")
    exit()


@cli.command()
@query_options
@click.option(
    "--csv",
    type=click.File("wb"),
    required=False,
    help="Provide filepath for dataframe output to csv",
)
@click.option(
    "--explain",
    is_flag=True,
    required=False,
    default=False,
    help="In the query output, show the query that was executed against the data warehouse",
)
@click.option(
    "--show-dataflow-plan",
    is_flag=True,
    required=False,
    default=False,
    help="Display dataflow plan in explain output",
)
@click.option(
    "--display-plans",
    is_flag=True,
    required=False,
    help="Display plans (e.g. metric dataflow) in the browser",
)
@click.option(
    "--decimals",
    required=False,
    default=2,
    help="Choose the number of decimal places to round for the numerical values",
)
@click.option(
    "--show-sql-descriptions",
    is_flag=True,
    default=False,
    help="Shows inline descriptions of nodes in displayed SQL",
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def query(
    cfg: CLIContext,
    metrics: List[str],
    group_by: List[str] = [],
    where: Optional[str] = None,
    start_time: Optional[dt.datetime] = None,
    end_time: Optional[dt.datetime] = None,
    order: Optional[List[str]] = None,
    limit: Optional[int] = None,
    csv: Optional[click.utils.LazyFile] = None,
    explain: bool = False,
    show_dataflow_plan: bool = False,
    display_plans: bool = False,
    decimals: int = DEFAULT_RESULT_DECIMAL_PLACES,
    show_sql_descriptions: bool = False,
) -> None:
    """Create a new query with MetricFlow and assembles a MetricFlowQueryResult."""
    start = time.time()
    spinner = Halo(text="Initiating query…", spinner="dots")
    spinner.start()
    mf_request = MetricFlowQueryRequest.create_with_random_request_id(
        metric_names=metrics,
        group_by_names=group_by,
        limit=limit,
        time_constraint_start=start_time,
        time_constraint_end=end_time,
        where_constraint=where,
        order_by_names=order,
    )

    explain_result: Optional[MetricFlowExplainResult] = None
    query_result: Optional[MetricFlowQueryResult] = None

    if explain:
        explain_result = cfg.mf.explain(mf_request=mf_request)
    else:
        query_result = cfg.mf.query(mf_request=mf_request)

    spinner.succeed(f"Success 🦄 - query completed after {time.time() - start:.2f} seconds")

    if explain:
        assert explain_result
        sql = (
            explain_result.rendered_sql_without_descriptions.sql_query
            if not show_sql_descriptions
            else explain_result.rendered_sql.sql_query
        )
        if show_dataflow_plan:
            click.echo("🔎 Generated Dataflow Plan + SQL (remove --explain to see data):")
            click.echo(
                textwrap.indent(
                    jinja2.Template(
                        textwrap.dedent(
                            """\
                            Metric Dataflow Plan:
                                {{ plan_text | indent(4) }}
                            """
                        ),
                        undefined=jinja2.StrictUndefined,
                    ).render(plan_text=dataflow_plan_as_text(explain_result.dataflow_plan)),
                    prefix="-- ",
                )
            )
            click.echo("")
        else:
            click.echo(
                "🔎 SQL (remove --explain to see data or add --show-dataflow-plan to see the generated dataflow plan):"
            )
        click.echo(sql)
        if display_plans:
            click.echo("Creating temporary directory for storing visualization output.")
            temp_path = tempfile.mkdtemp()
            svg_path = display_dag_as_svg(explain_result.dataflow_plan, temp_path)
            click.echo("")
            click.echo(f"Plan SVG saved to: {svg_path}")
        exit()

    assert query_result
    df = query_result.result_df
    # Show the data if returned successfully
    if df is not None:
        if df.empty:
            click.echo("🕳 Successful MQL query returned an empty result set.")
        elif csv is not None:
            # csv is a LazyFile that is file-like that works in this case.
            df.to_csv(csv, index=False)  # type: ignore
            click.echo(f"🖨 Successfully written query output to {csv.name}")
        else:
            # NOTE: remove `to_string` if no pandas dependency is < 1.1.0
            if parse(pd.__version__) >= parse("1.1.0"):
                click.echo(df.to_markdown(index=False, floatfmt=f".{decimals}f"))
            else:
                click.echo(df.to_string(index=False, float_format=lambda x: format(x, f".{decimals}f")))

        if display_plans:
            temp_path = tempfile.mkdtemp()
            svg_path = display_dag_as_svg(query_result.dataflow_plan, temp_path)
            click.echo(f"Plan SVG saved to: {svg_path}")


@cli.group()
@pass_config
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def list(cfg: CLIContext) -> None:  # noqa: D
    """Retrieve metadata values about metrics/dimensions/entities/dimension values."""


@list.command()
@click.option("--search", required=False, type=str, help="Filter available metrics by this search term")
@click.option(
    "--show-all-dimensions", is_flag=True, default=False, help="Show all dimensions associated with a metric."
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def metrics(cfg: CLIContext, show_all_dimensions: bool = False, search: Optional[str] = None) -> None:
    """List the metrics with their available dimensions.

    Automatically truncates long lists of dimensions, pass --show-all-dims to see all.
    """
    spinner = Halo(text="🔍 Looking for all available metrics...", spinner="dots")
    spinner.start()

    metrics = cfg.mf.list_metrics()

    if not metrics:
        spinner.fail("List of metrics unavailable.")

    filter_msg = ""
    if search is not None:
        num_metrics = len(metrics)
        metrics = [m for m in metrics if search.lower() in m.name.lower()]
        filter_msg = f" matching `{search}`, of a total of {num_metrics} available"

    spinner.succeed(f"🌱 We've found {len(metrics)} metrics{filter_msg}.")
    click.echo('The list below shows metrics in the format of "metric_name: list of available dimensions"')
    num_dims_to_show = MAX_LIST_OBJECT_ELEMENTS
    for m in metrics:
        # sort dimensions by whether they're local first(if / then global else local) then the dim name
        dimensions = sorted([dimension.granularity_free_qualified_name for dimension in m.dimensions])
        if show_all_dimensions:
            num_dims_to_show = len(dimensions)
        click.echo(
            f"• {click.style(m.name, bold=True, fg='green')}: {', '.join(dimensions[:num_dims_to_show])}"
            + (f" and {len(dimensions) - num_dims_to_show} more" if len(dimensions) > num_dims_to_show else "")
        )


@list.command()
@click.option(
    "--metrics",
    type=click_custom.SequenceParamType(min_length=1),
    default="",
    help="List dimensions by given metrics (intersection). Ex. --metrics bookings,messages",
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def dimensions(cfg: CLIContext, metrics: List[str]) -> None:
    """List all unique dimensions."""
    spinner = Halo(
        text="🔍 Looking for all available dimensions...",
        spinner="dots",
    )
    spinner.start()

    dimensions = cfg.mf.simple_dimensions_for_metrics(metrics)
    if not dimensions:
        spinner.fail("List of dimensions unavailable.")

    spinner.succeed(f"🌱 We've found {len(dimensions)} common dimensions for metrics {metrics}.")
    for dimension in dimensions:
        click.echo(f"• {click.style(dimension.granularity_free_qualified_name, bold=True, fg='green')}")


@list.command()
@click.option(
    "--metrics",
    type=click_custom.SequenceParamType(min_length=1),
    default="",
    help="List entities by given metrics (intersection). Ex. --metrics bookings,messages",
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def entities(cfg: CLIContext, metrics: List[str]) -> None:
    """List all unique entities."""
    spinner = Halo(
        text="🔍 Looking for all available entities...",
        spinner="dots",
    )
    spinner.start()

    entities = cfg.mf.entities_for_metrics(metrics)
    if not entities:
        spinner.fail("List of entities unavailable.")

    spinner.succeed(f"🌱 We've found {len(entities)} common entities for metrics {metrics}.")
    for entity in entities:
        click.echo(f"• {click.style(entity.name, bold=True, fg='green')}")


@cli.command()
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def health_checks(cfg: CLIContext) -> None:
    """Performs a health check against the DW provided in the configs."""
    spinner = Halo(
        text="🏥 Running health checks against your data warehouse... (This should not take longer than 30s for a successful connection)",
        spinner="dots",
    )
    spinner.start()
    res = cfg.run_health_checks()
    spinner.succeed("Health checks completed.")
    for test in res:
        test_res = res[test]
        if test_res["status"] != "SUCCESS":
            click.echo(f"• ❌ {click.style(test, bold=True, fg=('red'))}:  Failed with - {test_res['error_message']}.")
        else:
            click.echo(f"• ✅ {click.style(test, bold=True, fg=('green'))}: Success!")


@list.command()
@click.option("--dimension", required=True, type=str, help="Dimension to query values from")
@click.option(
    "--metrics",
    required=True,
    type=click_custom.SequenceParamType(min_length=1),
    help="Metrics that are associated with the dimension",
)
@start_end_time_options
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def dimension_values(
    cfg: CLIContext,
    metrics: List[str],
    dimension: str,
    start_time: Optional[dt.datetime] = None,
    end_time: Optional[dt.datetime] = None,
) -> None:
    """List all dimension values with the corresponding metrics."""
    spinner = Halo(
        text=f"🔍 Retrieving dimension values for dimension '{dimension}' of metrics '{', '.join(metrics)}'...",
        spinner="dots",
    )
    spinner.start()

    dim_vals: Optional[List[str]] = None

    try:
        dim_vals = cfg.mf.get_dimension_values(
            metric_names=metrics,
            get_group_by_values=dimension,
            time_constraint_start=start_time,
            time_constraint_end=end_time,
        )
    except Exception as e:
        spinner.fail()
        click.echo(
            textwrap.dedent(
                f"""\
                ❌ Failed to query dimension values for dimension {dimension} of metrics {', '.join(metrics)}.
                    ERROR: {str(e)}
                """
            )
        )
        exit(1)

    assert dim_vals
    spinner.succeed(
        f"🌱 We've found {len(dim_vals)} dimension values for dimension {dimension} of metrics {', '.join(metrics)}."
    )
    for dim_val in dim_vals:
        click.echo(f"• {click.style(dim_val, bold=True, fg='green')}")


def _print_issues(
    issues: SemanticManifestValidationResults, show_non_blocking: bool = False, verbose: bool = False
) -> None:  # noqa: D
    for issue in issues.errors:
        print(f"• {issue.as_cli_formatted_str(verbose=verbose)}")
    if show_non_blocking:
        for issue in issues.future_errors:
            print(f"• {issue.as_cli_formatted_str(verbose=verbose)}")
        for issue in issues.warnings:
            print(f"• {issue.as_cli_formatted_str(verbose=verbose)}")


def _run_dw_validations(
    validation_func: Callable[[SemanticManifest, Optional[int]], SemanticManifestValidationResults],
    validation_type: str,
    manifest: SemanticManifest,
    timeout: Optional[int],
) -> SemanticManifestValidationResults:
    """Helper handles the calling of data warehouse issue generating functions."""
    spinner = Halo(text=f"Validating {validation_type} against data warehouse...", spinner="dots")
    spinner.start()

    results = validation_func(manifest, timeout)
    if not results.has_blocking_issues:
        spinner.succeed(f"🎉 Successfully validated {validation_type} against data warehouse ({results.summary()})")
    else:
        spinner.fail(
            f"Breaking issues found when validating {validation_type} against data warehouse ({results.summary()})"
        )
    return results


def _data_warehouse_validations_runner(
    dw_validator: DataWarehouseModelValidator, manifest: SemanticManifest, timeout: Optional[int]
) -> SemanticManifestValidationResults:
    """Helper which calls the individual data warehouse validations to run and prints collected issues."""
    semantic_model_results = _run_dw_validations(
        dw_validator.validate_semantic_models, manifest=manifest, validation_type="semantic models", timeout=timeout
    )
    dimension_results = _run_dw_validations(
        dw_validator.validate_dimensions, manifest=manifest, validation_type="dimensions", timeout=timeout
    )
    entity_results = _run_dw_validations(
        dw_validator.validate_entities, manifest=manifest, validation_type="entities", timeout=timeout
    )
    measure_results = _run_dw_validations(
        dw_validator.validate_measures, manifest=manifest, validation_type="measures", timeout=timeout
    )
    metric_results = _run_dw_validations(
        dw_validator.validate_metrics, manifest=manifest, validation_type="metrics", timeout=timeout
    )

    return SemanticManifestValidationResults.merge(
        [semantic_model_results, dimension_results, entity_results, measure_results, metric_results]
    )


@cli.command()
@click.option(
    "--dw-timeout", required=False, type=int, help="Optional timeout for data warehouse validation steps. Default None."
)
@click.option(
    "--skip-dw",
    is_flag=True,
    default=False,
    help="If specified, skips the data warehouse validations",
)
@click.option("--show-all", is_flag=True, default=False, help="If specified, prints warnings and future-errors")
@click.option(
    "--verbose-issues", is_flag=True, default=False, help="If specified, prints any extra details issues might have"
)
@click.option(
    "--semantic-validation-workers",
    required=False,
    type=int,
    default=1,
    help="Optional. Uses the number of workers specified to run the semantic validations. Should only be used for exceptionally large configs",
)
@pass_config
@exception_handler
@log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
def validate_configs(
    cfg: CLIContext,
    dw_timeout: Optional[int] = None,
    skip_dw: bool = False,
    show_all: bool = False,
    verbose_issues: bool = False,
    semantic_validation_workers: int = 1,
) -> None:
    """Perform validations against the defined model configurations."""
    cfg.verbose = True

    if not show_all:
        print("(To see warnings and future-errors, run again with flag `--show-all`)")

    # Parsing Validation
    parsing_spinner = Halo(text="Building manifest from dbt project root", spinner="dots")
    parsing_spinner.start()
    project_root = pathlib.Path.cwd()

    try:
        semantic_manifest = dbtArtifacts.build_semantic_manifest_from_dbt_project_root(project_root=project_root)
        parsing_spinner.succeed("🎉 Successfully parsed manifest from dbt project")
    except Exception as e:
        parsing_spinner.fail(f"Exception found when parsing manifest from dbt project ({str(e)})")
        return

    # Semantic validation
    semantic_spinner = Halo(text="Validating semantics of built manifest", spinner="dots")
    semantic_spinner.start()
    model_issues = SemanticManifestValidator[SemanticManifest](
        max_workers=semantic_validation_workers
    ).validate_semantic_manifest(semantic_manifest)

    if not model_issues.has_blocking_issues:
        semantic_spinner.succeed(f"🎉 Successfully validated the semantics of built manifest ({model_issues.summary()})")
    else:
        semantic_spinner.fail(
            f"Breaking issues found when checking semantics of built manifest ({model_issues.summary()})"
        )
        _print_issues(model_issues, show_non_blocking=show_all, verbose=verbose_issues)
        return

    dw_results = SemanticManifestValidationResults()
    if not skip_dw:
        # fetch dbt adapters. This rebuilds the manifest again, but whatever.
        dw_validator = DataWarehouseModelValidator(sql_client=cfg.sql_client)
        dw_results = _data_warehouse_validations_runner(
            dw_validator=dw_validator, manifest=semantic_manifest, timeout=dw_timeout
        )

    merged_results = SemanticManifestValidationResults.merge([model_issues, dw_results])
    _print_issues(merged_results, show_non_blocking=show_all, verbose=verbose_issues)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import logging
from collections import defaultdict
from copy import deepcopy
from typing import Dict, List, Optional, Sequence, Set

from dbt_semantic_interfaces.pretty_print import pformat_big_objects
from dbt_semantic_interfaces.protocols.dimension import Dimension
from dbt_semantic_interfaces.protocols.entity import Entity
from dbt_semantic_interfaces.protocols.measure import Measure
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    DimensionReference,
    EntityReference,
    LinkableElementReference,
    MeasureReference,
    SemanticModelElementReference,
    SemanticModelReference,
    TimeDimensionReference,
)
from dbt_semantic_interfaces.type_enums import DimensionType, EntityType
from dbt_semantic_interfaces.type_enums.aggregation_type import AggregationType
from typing_extensions import override

from metricflow.errors.errors import InvalidSemanticModelError
from metricflow.model.semantics.element_group import ElementGrouper
from metricflow.model.spec_converters import MeasureConverter
from metricflow.protocols.semantics import SemanticModelAccessor
from metricflow.specs.specs import MeasureSpec, NonAdditiveDimensionSpec

logger = logging.getLogger(__name__)


class SemanticModelLookup(SemanticModelAccessor):
    """Tracks semantic information for semantic model held in a set of SemanticModelContainers.

    This implements both the SemanticModelAccessors protocol, the interface type we use throughout the codebase.
    That interface prevents unwanted calls to methods for adding semantic models to the container.
    """

    def __init__(  # noqa: D
        self,
        model: SemanticManifest,
    ) -> None:
        self._model = model
        self._measure_index: Dict[MeasureReference, List[SemanticModel]] = defaultdict(list)
        self._measure_aggs: Dict[
            MeasureReference, AggregationType
        ] = {}  # maps measures to their one consistent aggregation
        self._measure_agg_time_dimension: Dict[MeasureReference, TimeDimensionReference] = {}
        self._measure_non_additive_dimension_specs: Dict[MeasureReference, NonAdditiveDimensionSpec] = {}
        self._dimension_index: Dict[DimensionReference, List[SemanticModel]] = defaultdict(list)
        self._linkable_reference_index: Dict[LinkableElementReference, List[SemanticModel]] = defaultdict(list)
        self._entity_index: Dict[Optional[str], List[SemanticModel]] = defaultdict(list)
        self._entity_ref_to_entity: Dict[EntityReference, Optional[str]] = {}
        self._semantic_model_names: Set[str] = set()

        self._semantic_model_to_aggregation_time_dimensions: Dict[
            SemanticModelReference, ElementGrouper[TimeDimensionReference, MeasureSpec]
        ] = {}

        self._semantic_model_reference_to_semantic_model: Dict[SemanticModelReference, SemanticModel] = {}
        for semantic_model in self._model.semantic_models:
            self._add_semantic_model(semantic_model)

    def get_dimension_references(self) -> Sequence[DimensionReference]:  # noqa: D
        return tuple(self._dimension_index.keys())

    @staticmethod
    def get_dimension_from_semantic_model(
        semantic_model: SemanticModel, dimension_reference: LinkableElementReference
    ) -> Dimension:
        """Get dimension from semantic model."""
        for dim in semantic_model.dimensions:
            if dim.reference == dimension_reference:
                return dim
        raise ValueError(
            f"No dimension with name ({dimension_reference}) in semantic_model with name ({semantic_model.name})"
        )

    def get_dimension(self, dimension_reference: DimensionReference) -> Dimension:
        """Retrieves a full dimension object by name."""
        for dimension_source in self._dimension_index[dimension_reference]:
            dimension = SemanticModelLookup.get_dimension_from_semantic_model(
                semantic_model=dimension_source, dimension_reference=dimension_reference
            )

            return deepcopy(dimension)

        raise ValueError(
            f"Could not find dimension with name ({dimension_reference.element_name}) in configured semantic models"
        )

    def get_time_dimension(self, time_dimension_reference: TimeDimensionReference) -> Dimension:
        """Retrieves a full dimension object by name."""
        dimension_reference = time_dimension_reference.dimension_reference()

        if dimension_reference not in self._dimension_index:
            raise ValueError(
                f"Could not find dimension with name ({dimension_reference.element_name}) in configured semantic models"
            )

        for dimension_source in self._dimension_index[dimension_reference]:
            dimension = SemanticModelLookup.get_dimension_from_semantic_model(
                semantic_model=dimension_source, dimension_reference=dimension_reference
            )
            # TODO: Unclear if the deepcopy is necessary.
            return deepcopy(dimension)

        assert False, f"{time_dimension_reference} should have been in the dimension index"

    @property
    def measure_references(self) -> Sequence[MeasureReference]:  # noqa: D
        return list(self._measure_index.keys())

    @property
    def non_additive_dimension_specs_by_measure(self) -> Dict[MeasureReference, NonAdditiveDimensionSpec]:  # noqa: D
        return self._measure_non_additive_dimension_specs

    @staticmethod
    def get_measure_from_semantic_model(semantic_model: SemanticModel, measure_reference: MeasureReference) -> Measure:
        """Get measure from semantic model."""
        for measure in semantic_model.measures:
            if measure.reference == measure_reference:
                return measure

        raise ValueError(
            f"No dimension with name ({measure_reference.element_name}) in semantic_model with name ({semantic_model.name})"
        )

    def get_measure(self, measure_reference: MeasureReference) -> Measure:  # noqa: D
        if measure_reference not in self._measure_index:
            raise ValueError(f"Could not find measure with name ({measure_reference}) in configured semantic models")

        assert len(self._measure_index[measure_reference]) >= 1
        # Measures should be consistent across semantic models, so just use the first one.
        semantic_model = list(self._measure_index[measure_reference])[0]
        return SemanticModelLookup.get_measure_from_semantic_model(
            semantic_model=semantic_model, measure_reference=measure_reference
        )

    def get_entity_references(self) -> Sequence[EntityReference]:  # noqa: D
        return list(self._entity_ref_to_entity.keys())

    # DSC interface
    def get_semantic_models_for_measure(  # noqa: D
        self, measure_reference: MeasureReference
    ) -> Sequence[SemanticModel]:
        return self._measure_index[measure_reference]

    def get_agg_time_dimension_for_measure(  # noqa: D
        self, measure_reference: MeasureReference
    ) -> TimeDimensionReference:
        return self._measure_agg_time_dimension[measure_reference]

    def get_entity_in_semantic_model(self, ref: SemanticModelElementReference) -> Optional[Entity]:  # Noqa: d
        semantic_model = self.get_by_reference(ref.semantic_model_reference)
        if not semantic_model:
            return None

        for entity in semantic_model.entities:
            if entity.reference.element_name == ref.element_name:
                return entity

        return None

    def get_by_reference(self, semantic_model_reference: SemanticModelReference) -> Optional[SemanticModel]:  # noqa: D
        return self._semantic_model_reference_to_semantic_model.get(semantic_model_reference)

    def _add_semantic_model(self, semantic_model: SemanticModel) -> None:
        """Add semantic model semantic information, validating consistency with existing semantic models."""
        errors = []

        if semantic_model.reference in self._semantic_model_reference_to_semantic_model:
            errors.append(f"Semantic model {semantic_model.reference} already added.")

        for measure in semantic_model.measures:
            if measure.reference in self._measure_aggs and self._measure_aggs[measure.reference] != measure.agg:
                errors.append(
                    f"Conflicting aggregation (agg) for measure {measure.reference}. Currently registered as "
                    f"{self._measure_aggs[measure.reference]} but got {measure.agg}."
                )

        if len(errors) > 0:
            raise InvalidSemanticModelError(f"Error adding {semantic_model.reference}. Got errors: {errors}")

        self._semantic_model_to_aggregation_time_dimensions[semantic_model.reference] = ElementGrouper[
            TimeDimensionReference, MeasureSpec
        ]()

        for measure in semantic_model.measures:
            self._measure_aggs[measure.reference] = measure.agg
            self._measure_index[measure.reference].append(semantic_model)
            agg_time_dimension_reference = semantic_model.checked_agg_time_dimension_for_measure(measure.reference)

            matching_dimensions = tuple(
                dimension
                for dimension in semantic_model.dimensions
                if dimension.type == DimensionType.TIME
                and dimension.time_dimension_reference == agg_time_dimension_reference
            )

            matching_dimensions_count = len(matching_dimensions)
            if matching_dimensions_count != 1:
                raise RuntimeError(
                    f"Found {matching_dimensions_count} matching dimensions for {agg_time_dimension_reference} "
                    f"in {semantic_model}"
                )
            agg_time_dimension = matching_dimensions[0]
            if agg_time_dimension.type_params is None or agg_time_dimension.type_params.time_granularity is None:
                raise RuntimeError(
                    f"Aggregation time dimension does not have a time granularity set: {agg_time_dimension}"
                )

            primary_entity = SemanticModelLookup._resolved_primary_entity(semantic_model)

            if primary_entity is None:
                raise RuntimeError(
                    f"The semantic model should have a primary entity since there are dimensions, but it does not. "
                    f"Semantic model is:\n{pformat_big_objects(semantic_model)}"
                )

            self._semantic_model_to_aggregation_time_dimensions[semantic_model.reference].add_value(
                key=TimeDimensionReference(
                    element_name=agg_time_dimension.name,
                ),
                value=MeasureConverter.convert_to_measure_spec(measure=measure),
            )
            self._measure_agg_time_dimension[measure.reference] = TimeDimensionReference(
                element_name=agg_time_dimension.name
            )

            if measure.non_additive_dimension:
                non_additive_dimension_spec = NonAdditiveDimensionSpec(
                    name=measure.non_additive_dimension.name,
                    window_choice=measure.non_additive_dimension.window_choice,
                    window_groupings=tuple(measure.non_additive_dimension.window_groupings),
                )
                self._measure_non_additive_dimension_specs[measure.reference] = non_additive_dimension_spec
        for dim in semantic_model.dimensions:
            self._linkable_reference_index[dim.reference].append(semantic_model)
            self._dimension_index[dim.reference].append(semantic_model)
        for entity in semantic_model.entities:
            self._entity_ref_to_entity[entity.reference] = entity.name
            self._entity_index[entity.name].append(semantic_model)
            self._linkable_reference_index[entity.reference].append(semantic_model)

        self._semantic_model_reference_to_semantic_model[semantic_model.reference] = semantic_model

    @property
    def semantic_model_references(self) -> Sequence[SemanticModelReference]:  # noqa: D
        semantic_model_names_sorted = sorted(self._semantic_model_names)
        return tuple(SemanticModelReference(semantic_model_name=x) for x in semantic_model_names_sorted)

    def get_aggregation_time_dimensions_with_measures(
        self, semantic_model_reference: SemanticModelReference
    ) -> ElementGrouper[TimeDimensionReference, MeasureSpec]:
        """Return all time dimensions in a semantic model with their associated measures."""
        assert (
            semantic_model_reference in self._semantic_model_to_aggregation_time_dimensions
        ), f"Semantic Model {semantic_model_reference} is not known"
        return self._semantic_model_to_aggregation_time_dimensions[semantic_model_reference]

    def get_semantic_models_for_entity(self, entity_reference: EntityReference) -> Set[SemanticModel]:
        """Return all semantic models associated with an entity reference."""
        entity = self._entity_ref_to_entity[entity_reference]
        return set(self._entity_index[entity])

    @staticmethod
    def get_entity_from_semantic_model(
        semantic_model: SemanticModel, entity_reference: LinkableElementReference
    ) -> Entity:
        """Get entity from semantic model."""
        for entity in semantic_model.entities:
            if entity.reference == entity_reference:
                return entity

        raise ValueError(
            f"No entity with name ({entity_reference}) in semantic_model with name ({semantic_model.name})"
        )

    @staticmethod
    def _resolved_primary_entity(semantic_model: SemanticModel) -> Optional[EntityReference]:
        """Return the primary entity for dimensions in the model."""
        primary_entity_reference = semantic_model.primary_entity_reference

        entities_with_type_primary = tuple(
            entity for entity in semantic_model.entities if entity.type == EntityType.PRIMARY
        )

        # This should be caught by the validation, but adding a sanity check.
        assert len(entities_with_type_primary) <= 1, f"Found >1 primary entity in {semantic_model}"
        if primary_entity_reference is not None:
            assert len(entities_with_type_primary) == 0, (
                f"The primary_entity field was set to {primary_entity_reference}, but there are non-zero entities with "
                f"type {EntityType.PRIMARY} in {semantic_model}"
            )

        if primary_entity_reference is not None:
            return primary_entity_reference

        if len(entities_with_type_primary) > 0:
            return entities_with_type_primary[0].reference

        return None

    @staticmethod
    @override
    def entity_links_for_local_elements(semantic_model: SemanticModel) -> Sequence[EntityReference]:
        primary_entity_reference = semantic_model.primary_entity_reference

        possible_entity_links = set()
        if primary_entity_reference is not None:
            possible_entity_links.add(primary_entity_reference)

        for entity in semantic_model.entities:
            if entity.is_linkable_entity_type:
                possible_entity_links.add(entity.reference)

        return sorted(possible_entity_links, key=lambda entity_reference: entity_reference.element_name)



from __future__ import annotations

import datetime
import logging
import os
import warnings
from dataclasses import dataclass

import _pytest.config
import pytest
import sqlalchemy.util
from _pytest.fixtures import FixtureRequest
from sqlalchemy.engine import make_url

from metricflow.random_id import random_id
from metricflow.test.fixtures.sql_clients.common_client import SqlDialect
from metricflow.test.table_snapshot.table_snapshots import SqlTableSnapshotHash, SqlTableSnapshotRepository

logger = logging.getLogger(__name__)


@dataclass
class MetricFlowTestSessionState:
    """State that is shared between tests during a testing session."""

    sql_engine_url: str
    sql_engine_password: str
    # Where MF system tables should be stored.
    mf_system_schema: str
    # Where tables for test data sets should be stored.
    mf_source_schema: str

    # Number of plans that were displayed to the user.
    display_plans: bool
    # Whether to overwrite any text files that were generated.
    overwrite_snapshots: bool
    # Number of plans that were displayed to the user.
    plans_displayed: int
    # Maximum number of plans to display to the user. If this is exceeded, an exception should be thrown. This is to
    # help avoid a case where an excessive number of plans are displayed. This could happen if the user accidentally
    # runs all tests when they were just looking to run one test and visualize the associated plan.
    max_plans_displayed: int

    # The source schema contains tables that are used for running tests. If this is set, a source schema in the SQL
    # is created and persisted between runs. The source schema name includes a hash of the tables that should be in
    # the schema, so
    use_persistent_source_schema: bool


DISPLAY_PLANS_CLI_FLAG = "--display-plans"
OVERWRITE_SNAPSHOTS_CLI_FLAG = "--overwrite-snapshots"
USE_PERSISTENT_SOURCE_SCHEMA_CLI_FLAG = "--use-persistent-source-schema"


def pytest_addoption(parser: _pytest.config.argparsing.Parser) -> None:
    """Add options for running pytest through the CLI."""
    parser.addoption(DISPLAY_PLANS_CLI_FLAG, action="store_true", help="Displays plans as SVGs in a browser tab if set")
    parser.addoption(
        OVERWRITE_SNAPSHOTS_CLI_FLAG,
        action="store_true",
        help="Overwrites existing snapshots by ones generated during this testing session",
    )
    parser.addoption(
        USE_PERSISTENT_SOURCE_SCHEMA_CLI_FLAG,
        action="store_true",
        help="Use a source schema that is persisted between testing sessions. The name of the schema is generated from"
        "a hash of the source data, and the schema is created / populated if it does not exist.",
    )


@pytest.fixture(scope="session")
def mf_test_session_state(  # noqa: D
    request: FixtureRequest,
    disable_sql_alchemy_deprecation_warning: None,
    source_table_snapshot_repository: SqlTableSnapshotRepository,
) -> MetricFlowTestSessionState:
    engine_url = os.environ.get("MF_SQL_ENGINE_URL")
    assert engine_url is not None, (
        "MF_SQL_ENGINE_URL environment variable has not been set! Are you running in a properly configured "
        "environment? Check out our CONTRIBUTING.md for pointers to our environment configurations."
    )
    engine_password = os.environ.get("MF_SQL_ENGINE_PASSWORD", "")

    current_time = datetime.datetime.now().strftime("%Y_%m_%d")
    random_suffix = random_id()
    mf_system_schema = f"mf_test_{current_time}_{random_suffix}"
    default_source_schema = mf_system_schema

    use_persistent_source_schema = bool(request.config.getoption(USE_PERSISTENT_SOURCE_SCHEMA_CLI_FLAG, default=False))
    if use_persistent_source_schema:
        source_table_snapshots_hash = SqlTableSnapshotHash.create_from_hashes(
            tuple(table_snapshot.snapshot_hash for table_snapshot in source_table_snapshot_repository.table_snapshots)
        )
        # Since the source schema is used in the SQL output snapshots (but replaced with *'s), make the name of the
        # schema to be the same length so that the SQL output snapshots don't change.
        persistent_source_schema_prefix = "mf_test_src_"

        # Ensure that there are a reasonable number of hash digits in the schema so that we don't get collisions.
        available_hash_digits = max(len(default_source_schema) - len(persistent_source_schema_prefix), 0)
        if available_hash_digits < 8:
            raise RuntimeError(
                f"The generated name for the persistent source schema would have {available_hash_digits} data hash "
                f"digits, which is not enough to have low collisions. default_source_schema: {default_source_schema} "
                f"source_schema_prefix: {persistent_source_schema_prefix}"
            )

        # Use as many digits of the hash without exceeding the length of the default source schema.
        mf_source_schema = f"{persistent_source_schema_prefix}{source_table_snapshots_hash.str_value}"[
            : len(default_source_schema)
        ]

        if len(mf_source_schema) != len(default_source_schema):
            raise RuntimeError(
                f"The persistent source schema should be the same length as the default source schema to keep the "
                f"generated SQL snapshots consistent between non-persistent and persistent modes. persistent source "
                f"schema: {mf_source_schema} default source schema: {default_source_schema}"
            )

        logger.info(
            f"Since the flag {USE_PERSISTENT_SOURCE_SCHEMA_CLI_FLAG} was specified, this session will use the "
            f"persistent source schema {mf_source_schema}. If the required source tables do not exist in this "
            f"schema, they will be created. However, the source schema (and the associated tables) will not "
            f"be dropped at the end of the testing session."
        )
    else:
        mf_source_schema = default_source_schema

    return MetricFlowTestSessionState(
        sql_engine_url=engine_url,
        sql_engine_password=engine_password,
        mf_system_schema=mf_system_schema,
        mf_source_schema=mf_source_schema,
        display_plans=bool(request.config.getoption(DISPLAY_PLANS_CLI_FLAG, default=False)),
        overwrite_snapshots=bool(request.config.getoption(OVERWRITE_SNAPSHOTS_CLI_FLAG, default=False)),
        plans_displayed=0,
        max_plans_displayed=6,
        use_persistent_source_schema=bool(
            request.config.getoption(USE_PERSISTENT_SOURCE_SCHEMA_CLI_FLAG, default=False)
        ),
    )


def dialect_from_url(url: str) -> SqlDialect:
    """Return the SQL dialect specified in the URL in the configuration."""
    dialect_protocol = make_url(url.split(";")[0]).drivername.split("+")
    if len(dialect_protocol) > 2:
        raise ValueError(f"Invalid # of +'s in {url}")
    return SqlDialect(dialect_protocol[0])


@pytest.fixture(scope="session", autouse=True)
def warn_user_about_slow_tests_without_parallelism(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
) -> None:
    worker_count_env_var = os.environ.get("PYTEST_XDIST_WORKER_COUNT", "1")
    try:
        num_workers = int(worker_count_env_var)
    except ValueError as e:
        raise ValueError(
            f"Could not convert environment variable PYTEST_XDIST_WORKER_COUNT to int! "
            f"Value in environ was: {worker_count_env_var}"
        ) from e

    num_items = len(request.session.items)
    dialect = dialect_from_url(mf_test_session_state.sql_engine_url)

    # If already running in parallel or if there's not many test items, no need to print the warning. Picking 10/30 as
    # the thresholds, but not much thought has been put into it.
    if num_workers > 1 or num_items < 10:
        return

    # Since DuckDB / Postgres is fast, use 30 as the threshold.
    if (dialect is SqlDialect.DUCKDB or dialect is SqlDialect.POSTGRESQL) and num_items < 30:
        return

    if num_items > 10:
        warnings.warn(
            f"This test session with {dialect.name} and {num_items} item(s) is running with {num_workers} worker(s). "
            f'Consider using the pytest-xdist option "-n <number of workers>" to parallelize execution and speed '
            f"up the session."
        )


@pytest.fixture(scope="session", autouse=True)
def disable_sql_alchemy_deprecation_warning() -> None:
    """Since MF is tied to using SQLAlchemy 1.x.x due to the Snowflake connector, silence 2.0 deprecation warnings."""
    # Seeing 'error: Module has no attribute "SILENCE_UBER_WARNING"' in the type checker, but this seems to work.
    sqlalchemy.util.deprecations.SILENCE_UBER_WARNING = True  # type:ignore
